from __future__ import print_function, division
from SLOPpy.subroutines.common import *
from SLOPpy.subroutines.spectral_subroutines import *
from SLOPpy.subroutines.io_subroutines import *
from SLOPpy.subroutines.fit_subroutines import *
from SLOPpy.subroutines.plot_subroutines import *
from SLOPpy.subroutines.shortcuts import *
from SLOPpy.telluric_molecfit_v1_preparation import compute_telluric_molecfit_v1_preparation

__all__ = ["compute_telluric_molecfit_v1_coadd",
           "plot_telluric_molecfit_v1_coadd",
           "compute_telluric_molecfit_v1",
           "plot_telluric_molecfit_v1"]

subroutine_name = 'telluric_molecfit_v1_coadd'


def compute_telluric_molecfit_v1_coadd(config_in, no_coadding=False):
    """
    Lazy workaround
    :param config_in:
    :param kwargs:
    :return:
    """

    night_dict = from_config_get_nights(config_in)
    instrument_dict = from_config_get_instrument(config_in)
    molecfit_dict = from_config_get_molecfit(config_in)

    compute_telluric_molecfit_v1_preparation(config_in)

    if no_coadding==True:
        print()
        print("--- WARNING ---: MOLECFIT model will be computed on individual spectra rather than on coadded spectra ")
        print("                 This task is accomplished by the same routine that computes the model on coadded spectra")
        print("                 With the only difference that the coadded spectra is made by a single observations")
        print("                 This choice proves a lot the maintenance of the code. As a consequence you will see a lot")
        print("                 of 'coadded' messages while running the code. ")
        print()
        molecfit_dict['exptime_coadd'] = 0.001
        subroutine_name = 'telluric_molecfit_v1'
    else:
        subroutine_name = 'telluric_molecfit_coadd_v1'

    for night in night_dict:

        instrument_name = night_dict[night]['instrument']
        template_dict = instrument_dict[instrument_name]['telluric_template']

        try:
            telluric = load_from_cpickle('telluric', config_in['output'], night)
            print("{0:45s} Night:{1:15s}   {2:s}".format(subroutine_name, night, 'Retrieved'))
            continue
        except:
            print("{0:45s} Night:{1:15s}   {2:s}".format(subroutine_name, night, 'Computing'))
            print()

        print('    instrument :', instrument_name)
        print()

        tellprep = load_from_cpickle('telluric_molecfit_preparation', config_in['output'], night)

        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)

        """ Retrieving the observations"""
        calib_data = load_from_cpickle('calibration_fibA', config_in['output'], night)
        input_data = retrieve_observations(config_in['output'], night, lists['observations'], use_telluric=False)
        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

        processed = {
            'subroutine': 'telluric_molecfit',
            'n_orders': 0,
            'n_pixels': 0,
        }

        telluric = {
            'subroutine': 'telluric_molecfit',
            'reference_frame': 'observer'
        }

        processed['airmass_ref'] = 0.000
        processed['telluric'] = {}
        processed['rebin'] = {}
        processed['work_dir'] = tellprep['work_dir']

        """
        Molecfit works on pixel grid, so we must ensure that the spectra are rebinned always on the same wavelength
        scale and same wavelength step. We use local arrays for this purpose
        """

        processed['rebin']['wave'] = np.arange(input_data['coadd']['wavelength_range'][0],
                                               input_data['coadd']['wavelength_range'][1],
                                               molecfit_dict['rebinning_step'],
                                               dtype=np.double)

        processed['rebin']['size'] = np.size(processed['rebin']['wave'])
        processed['rebin']['step'] = np.ones(processed['rebin']['size'],
                                             dtype=np.double) * molecfit_dict['rebinning_step']

        processed['rebin'] = {
            'wave': input_data['coadd']['wave'],
            'size': input_data['coadd']['size'],
            'step': input_data['coadd']['step'],
        }

        n_coadd = 0
        n_reference = 0
        texp_cumulated = 0.00
        texp_total = 0.000
        coadd_list = []

        # Computing the total integration time
        for n_obs, obs in enumerate(lists['observations']):
            texp_total += input_data[obs]['EXPTIME']

        print('  Writing data and configuration files for molecfit+calctrans')
        print()

        # There must be a more elegant way to do this, but I'm, not aware of it
        for n_obs, obs in enumerate(lists['observations']):

            processed[obs] = {
                'n_orders': input_data[obs]['n_orders'],
                'n_pixels': input_data[obs]['n_pixels']
            }

            """ e2ds spectra are rescaled and then rebinned while keeping them in the Observer Reference Frame"""

            processed[obs]['e2ds_rescaling'], processed[obs]['e2ds_rescaled'], processed[obs]['e2ds_rescaled_err'] = \
                perform_rescaling(input_data[obs]['wave'],
                                  input_data[obs]['e2ds'],
                                  input_data[obs]['e2ds_err'],
                                  observational_pams['wavelength_rescaling'])

            preserve_flux = input_data[obs].get('absolute_flux', True)

            processed[obs]['rebin_ORF'] = \
                rebin_2d_to_1d(input_data[obs]['wave'],
                               input_data[obs]['step'],
                               processed[obs]['e2ds_rescaled'],
                               calib_data['blaze'],
                               processed['rebin']['wave'],
                               processed['rebin']['step'],
                               preserve_flux=preserve_flux,
                               rv_shift=0.00)

            """ Molecfit analysis is skipped if the telluric correction has been computed already"""
            # if os.path.isfile('./molecfit_'+night +'/output/'+obs+'_ORF_s1d_TAC.dat'):
            # print('    molecfit+calctrans results for ' + obs + ' already available')
            # continue

            """ the spectra is saved as an ASCII file in a format suitable for molecfit """
            fileout = open('./' + processed['work_dir'] + '/' + obs + '_ORF_s1d.dat', 'w')
            for w, f in zip(processed['rebin']['wave'], processed[obs]['rebin_ORF']):
                fileout.write('{0:12.6f} {1:12.6f} \n'.format(w, f))
            fileout.close()

            """
            processed[obs]['rebin_SRF'] = \
            rebin_2d_to_1d(input_data[obs]['wave'],
                           input_data[obs]['step'],
                           processed[obs]['e2ds_rescaled'],
                           calib_data['blaze'],
                           processed['rebin']['wave'],
                           processed['rebin']['step'],
                           preserve_flux=preserve_flux,
                           rv_shift = observational_pams[obs]['rv_shift_ORF2SRF'])

            fileout = open('./' + processed['work_dir'] + '/' + obs + '_SRF_s1d.dat','w')
            for w, f in zip(processed['rebin']['wave'], processed[obs]['rebin_SRF']):
                fileout.write('{0:12.6f} {1:12.6f} \n'.format(w, f))
            fileout.close()
            """

            """ spectra is coadded to increase the SNR of the spectrum analyzed by molecfit """
            if n_coadd == 0:

                reference_name = 'coadded_{0:03d}'.format(n_reference)

                rebin_coadd = processed[obs]['rebin_ORF'].copy()

                molecfit_pams = {
                    'MJD': input_data[obs]['MJD'],
                    'UTC': input_data[obs]['UTC'],
                    'ELEVATION': input_data[obs]['ELEVATION'],
                    'HUMIDITY': input_data[obs]['HUMIDITY'],
                    'PRESSURE': input_data[obs]['PRESSURE'],
                    'TEMPERATURE_EN': input_data[obs]['TEMPERATURE_EN'],
                    'TEMPERATURE_M1': input_data[obs]['TEMPERATURE_M1']}

                coadded_files = open('./' + processed['work_dir'] + '/' + reference_name + '_files.list', 'w')

                coadd_list.append(reference_name)

            else:
                rebin_coadd += processed[obs]['rebin_ORF']

                molecfit_pams['MJD'] += input_data[obs]['MJD']
                molecfit_pams['UTC'] += input_data[obs]['UTC']
                molecfit_pams['ELEVATION'] += input_data[obs]['ELEVATION']
                molecfit_pams['HUMIDITY'] += input_data[obs]['HUMIDITY']
                molecfit_pams['PRESSURE'] += input_data[obs]['PRESSURE']
                molecfit_pams['TEMPERATURE_EN'] += input_data[obs]['TEMPERATURE_EN']
                molecfit_pams['TEMPERATURE_M1'] += input_data[obs]['TEMPERATURE_M1']

            n_coadd += 1
            coadded_files.write(obs + '\n')

            texp_cumulated += input_data[obs]['EXPTIME']

            # TODO: input from configuration file for molecfit installation path
            bash_script = open('./' + processed['work_dir'] + '/molecfit_exec_' + obs + '.source', 'w')
            bash_script.write('#!/bin/bash \n')

            bash_script.write('export TMPDIR=$PWD\n')
            bash_script.write('echo  "   " executing calctrans on ' + obs + ' \n')
            bash_script.write(molecfit_dict['installation_path'] + 'calctrans ' +
                              obs + '.par > ' + obs + '_calctrans.log\n')
            bash_script.close()

            write_molecfit_v1_par('./' + processed['work_dir'] + '/' + obs + '.par',
                                  obs + '_ORF_s1d.dat',
                                  reference_name,
                                  'include_' + night + '.dat',
                                  input_data[obs]['molecfit'],
                                  input_data[obs])

            if (texp_cumulated >= molecfit_dict['exptime_coadd'] and
                    texp_total-texp_cumulated >= molecfit_dict['exptime_coadd']) \
                    or n_obs == len(lists['observations'])-1:

                coadded_files.close()
                print('   Coadded spectrum: ', n_reference)

                rebin_coadd /= n_coadd

                """ the spectra is saved as an ASCII file in a format suitable for molecfit """
                fileout = open('./' + processed['work_dir'] + '/' + reference_name + '_ORF_s1d.dat', 'w')
                for w, f in zip(processed['rebin']['wave'], rebin_coadd):
                    fileout.write('{0:12.6f} {1:12.6f} \n'.format(w, f))
                fileout.close()

                """ Average of the observational parameters """
                for key in molecfit_pams:
                    molecfit_pams[key] /= n_coadd

                molecfit_pams['GEOELEV'] = input_data[obs]['GEOELEV']
                molecfit_pams['GEOLONG'] = input_data[obs]['GEOLONG']
                molecfit_pams['GEOLAT'] = input_data[obs]['GEOLAT']

                # TODO: input from configuration file for molecfit installation path
                bash_script = open('./' + processed['work_dir'] + '/molecfit_exec_' + reference_name + '.source', 'w')
                bash_script.write('#!/bin/bash \n')
                bash_script.write('export TMPDIR=$PWD\n')

                bash_script.write('echo  "   " executing molecfit+calctrans on ' + reference_name + ' \n')
                bash_script.write(molecfit_dict['installation_path'] + 'molecfit ' +
                                  reference_name + '.par > ' + reference_name + '_molecfit.log\n')
                bash_script.write(molecfit_dict['installation_path'] + 'calctrans ' +
                                  reference_name + '.par > ' + reference_name + '_calctrans.log\n')
                bash_script.close()

                # TODO: cycle with variation in UTC until molecfit exits succesfully
                #
                # while True:
                #
                #   # write parameter file
                #   # execute molecfit
                #   # check if file _tac.fits has been written (= successful run)
                #   if cond:
                #       break
                #

                utc_reference = molecfit_pams['UTC'] * 1.
                utc_incremental = True
                utc_increase = 500.

                while True:

                    if os.path.exists('./' + processed['work_dir'] + '/output/' + reference_name + '_tac.asc'):
                        print('  molecfit for ' + reference_name + ' previously completed')
                        print()
                        break

                    write_molecfit_v1_par('./' + processed['work_dir'] + '/' + reference_name + '.par',
                                          reference_name + '_ORF_s1d.dat',
                                          reference_name,
                                          'include_' + night + '.dat',
                                          input_data[obs]['molecfit'],
                                          molecfit_pams)

                    os.system('cd ' + processed['work_dir'] + '/ && . ./molecfit_exec_' + reference_name + '.source')

                    if os.path.exists('./' + processed['work_dir'] + '/output/' + reference_name + '_tac.asc'):
                        print('  molecfit for ' + reference_name + ' successfully completed')
                        print()
                        break

                    if molecfit_pams['UTC'] > 86400 - utc_increase:
                        utc_incremental = False
                        molecfit_pams['UTC'] = utc_reference

                    if utc_incremental:
                        molecfit_pams['UTC'] += utc_increase
                        print('  molecfit for {0:s} crashed, UTC increased from {1:6.0f} to {2:6.0f} '.format(
                            reference_name, utc_reference, molecfit_pams['UTC']))
                    else:
                        molecfit_pams['UTC'] -= utc_increase
                        print('  molecfit for {0:s} crashed, UTC decreased from {1:6.0f} to {2:6.0f} '.format(
                            reference_name, utc_reference, molecfit_pams['UTC']))

                n_coadd = 0
                n_reference += 1
                texp_total -= texp_cumulated
                texp_cumulated = 0.0

        """
        Execute molecfit runs on all the coadded spectra
        """

        # for reference_name in coadd_list:
        #    os.system('cd  molecfit_' + night + '/ && . ./molecfit_exec_' + reference_name + '.source')
        #
        print()
        print('  molecfit completed')

        for obs in lists['observations']:

            if os.path.exists('./' + processed['work_dir'] + '/output/' + obs + '_ORF_s1d_TAC.dat'):
                print('  skipping calctrans execution for observation ' + obs)
            else:
                print('  calctrans execution for observation ' + obs)
                os.system('cd ' + processed['work_dir'] + '/ && . ./molecfit_exec_' + obs + '.source')

        print()
        print('  calctrans completed')

        for n_obs, obs in enumerate(lists['observations']):

            telluric[obs] = {}

            """ Loading the telluric spectrum from the output directory of molecfit """
            telluric_molecfit = np.genfromtxt(
                './' + processed['work_dir'] + '/output/'+obs+'_ORF_s1d_TAC.dat', usecols=2)
            """ rebinning onto the e2ds wave scale"""

            if molecfit_dict.get('fix_telluric', True):
                print('  fix_telluric applied -  temporary workaround for line at 5885.97 A [ORF]')
                line_boundaries = [5885.74, 5886.21]
                sel = (processed['rebin']['wave'] > line_boundaries[0]) \
                    & (processed['rebin']['wave'] < line_boundaries[1])
                tell_cont = np.amax(telluric_molecfit[sel])

                telluric_molecfit[sel] = (telluric_molecfit[sel] - tell_cont) / 2.0 + tell_cont

            telluric[obs]['spectrum'] = \
                rebin_1d_to_2d(processed['rebin']['wave'],
                               processed['rebin']['step'],
                               telluric_molecfit,
                               input_data[obs]['wave'],
                               input_data[obs]['step'],
                               preserve_flux=False)

            try:
                telluric[obs]['spectrum'] = np.nan_to_num(nan=1.0, posinf=1.0, neginf=1.0)
            except:
                temp = ~(np.isfinite(telluric[obs]['spectrum']))
                telluric[obs]['spectrum'][temp] = 1.0
            sel = telluric[obs]['spectrum'] < 0.0001
            telluric[obs]['spectrum'][sel] = 1.0

            telluric[obs]['airmass'] = input_data[obs]['AIRMASS']

            " for compatibilty to some plots, even if it doesn't make any sense"
            telluric[obs]['airmass_ref'] = 0.000
            telluric[obs]['spectrum_noairmass'] = np.power(telluric[obs]['spectrum'],
                                                           telluric[obs]['airmass_ref'] - input_data[obs]['AIRMASS'])
            telluric[obs]['null'] = telluric[obs]['spectrum_noairmass'] < 0.001
            telluric[obs]['spectrum_noairmass'][telluric[obs]['null']] = 1.0
            # we just copy the spectrum file, it's it's a model itself
            telluric[obs]['spline'] = telluric[obs]['spectrum'].copy()

            processed[obs]['e2ds_corrected'] = processed[obs]['e2ds_rescaled'] / telluric[obs]['spectrum']
            processed[obs]['e2ds_corrected_err'] = processed[obs]['e2ds_rescaled_err'] / telluric[obs]['spectrum']

        save_to_cpickle('telluric', telluric, config_in['output'], night)
        save_to_cpickle('telluric_processed', processed, config_in['output'], night)

        print()
        print("Night ", night, " completed")


def plot_telluric_molecfit_v1_coadd(config_in, night_input='', no_coadding=False):
    import matplotlib.pyplot as plt

    night_dict = from_config_get_nights(config_in)
    instrument_dict = from_config_get_instrument(config_in)
    system_dict = from_config_get_system(config_in)

    if night_input == '':
        night_list = night_dict
    else:
        night_list = np.atleast_1d(night_input)

    if no_coadding==True:
        print()
        print("--- WARNING ---: MOLECFIT model will be computed on individual spectra rather than on coadded spectra ")
        print("                 This task is accomplished by the same routine that computes the model on coadded spectra")
        print("                 With the only difference that the coadded spectra is made by a single observations")
        print("                 This choice proves a lot the maintenance of the code. As a consequence you will see a lot")
        print("                 of 'coadded' messages while running the code. ")
        print()
        subroutine_name = 'telluric_molecfit'
    else:
        subroutine_name = 'telluric_molecfit_coadd_v1'

    for night in night_list:

        # plt.scatter(rescaling_array, computed_std, c='C0', zorder=1)
        # plt.scatter(sel_factor, sel_stdev, c='C1', zorder=2)
        # plt.plot(rescaling_array, np.polyval(coeff, rescaling_array))
        # plt.plot(rescaling_array, 2*rescaling_array*coeff[0] + coeff[1] )
        # plt.plot()

        print("{0:45s} Night:{1:15s}   {2:s}".format(subroutine_name, night, 'Plotting'))

        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)
        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

        """ Retrieving the analysis"""
        try:
            processed = load_from_cpickle('telluric_processed', config_in['output'], night)
            telluric = load_from_cpickle('telluric', config_in['output'], night)
        except:
            print()
            print("No telluric correction, no plots")
            continue

        input_data = retrieve_observations(config_in['output'], night, lists['observations'], use_telluric=False)

        colors, cmap, line_colors = make_color_array(lists, observational_pams)

        fig = plt.figure(figsize=(12, 6))
        gs = GridSpec(2, 2, width_ratios=[50, 1])

        ax1 = plt.subplot(gs[0, 0])
        ax2 = plt.subplot(gs[1, 0], sharex=ax1)
        cbax1 = plt.subplot(gs[:, 1])

        lift_spectrum = 0.25

        for i, obs in enumerate(lists['observations']):
            color_array = cmap(i / len(lists['observations']))

            for order in range(0, processed[obs]['n_orders']):

                if order == 0 and i == 0:
                    ax1.plot(input_data[obs]['wave'][order, :],
                             processed[obs]['e2ds_rescaled'][order, :],
                             c=color_array, lw=1, alpha=0.5, label='uncorrected')
                    ax1.scatter(input_data[obs]['wave'][order, :],
                                processed[obs]['e2ds_corrected'][order, :],
                                s=1, c=np.atleast_2d(color_array), label='corrected')
                else:
                    ax1.plot(input_data[obs]['wave'][order, :],
                             processed[obs]['e2ds_rescaled'][order, :],
                             c=color_array, lw=1, alpha=0.5)
                    ax1.scatter(input_data[obs]['wave'][order, :],
                                processed[obs]['e2ds_corrected'][order, :],
                                s=1, c=np.atleast_2d(color_array))

                # ax1.plot(processed[obs]['wave'][order, :],
                #            e2ds_rescaled[order, :]+lift_spectrum,
                #            c=color_array, lw=1, alpha=0.5)
                # ax1.scatter(processed[obs]['wave'][order, :],
                #            e2ds_rescaled_corrected_spline[order, :]+lift_spectrum,
                #            s=1, c=np.atleast_2d(color_array))

                ax2.plot(input_data[obs]['wave'][order, :],
                         telluric[obs]['spectrum'][order, :],
                         c=color_array)
                ax2.axhline(1.00, c='k')

                # ax2.plot(processed[obs]['wave'][order, :],
                #         telluric[obs]['spline'][order, :]+lift_spectrum,
                #         c=color_array)
                # ax2.axhline(1.00+lift_spectrum, c='k')

        # ax2.plot(input_data['coadd']['wave'],telluric['stellarRF']['spline_eval']+0.1,c='k')
        # ax2.scatter(input_data['coadd']['wave'],telluric['stellarRF']['spectrum']+0.1,c='r', s=2)

        ax1.legend(loc=3)
        ax1.set_title('Night: ' + night)

        ax2.set_xlabel('$\lambda$ [$\AA$]')

        try:
            instrument = night_dict[night]['instrument']
            comparison_file = config_in['instruments'][instrument]['telluric_comparison']
            comparison_data = np.genfromtxt(comparison_file, skip_header=1)
            if comparison_data[0, 0] < 1000.0:
                nm2Ang = 10.
            else:
                nm2Ang = 1.
            ax1.plot(comparison_data[:, 0]*nm2Ang, comparison_data[:, 1], c='C0', zorder=1000)
            ax2.plot(comparison_data[:, 0]*nm2Ang, comparison_data[:, 1], c='C0', zorder=1000)
        except:
            pass

        sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(vmin=colors[0], vmax=colors[-1]))
        sm.set_array([])  # You have to set a dummy-array for this to work...
        cbar = plt.colorbar(sm, cax=cbax1)
        cbar.set_label('BJD - 2450000.0')
        fig.subplots_adjust(wspace=0.05, hspace=0.4)
        plt.show()


def compute_telluric_molecfit_v1(config_in):
    compute_telluric_molecfit_v1_coadd(config_in, no_coadding=True)

def plot_telluric_molecfit_v1(config_in, night_input=''):
    plot_telluric_molecfit_v1_coadd(config_in, night_input=night_input, no_coadding=True)



config_default = {
    'molecfit': {
        'installation_path': '/usr/local/eso/bin/',
        'include_telluric': './additional_files/include_regions_ORF.dat',
        'include_stellar': './additional_files/include_regions_SRF.dat',
        'exptime_coadd': 2600,
        'rebinning_step': 0.010
    },
    'instruments': {
        'wavelength_step': 0.01000000,
        'use_rv_from_ccf': False,
        'use_analytical_rvs': False,
        'linear_fit_method': 'linear_curve_fit',
        'orders': None,
        'telluric': None,
        'telluric_template': None
    },
    'refraction': {
        'approach': 'individual_order',
        'method': 'polynomial',
        'fit_order': 5,
        'fit_iters': 3,
        'fit_sigma': 5,
        'knots_spacing': 25.00,
        'reference_night': False,
        'reference_instrument': False
    },
    'master-out': {
        'wavelength_range': [5880.000, 5906.000],
        'wavelength_step': 0.010000000,
        'binning_factor': 20,
        'use_smoothed': False,
        'use_composite': False,
        'boxcar_smoothing': 3,
    },
    'pca_parameters': {
        'iterations': 5,
        'ref_iteration': 0,
    },
    'settings': {
        'full_output': False
    }
}

""" List ok keywords that are copied from the instrument dictionary
    when they are not explicitily specified
"""
copy_from_instrument = [
    'telluric_template',
    'telluric',
    'mask',
    'use_rv_from_ccf',
    'use_analytical_rvs',
    'telescope',
    'spectral_selection',
    'apply_ESO_telluric_correction',
    'use_ESO_sky_correction',
    'use_ESO_deblazed'
]



from __future__ import print_function, division
from SLOPpy.subroutines.common import *
from SLOPpy.subroutines.spectral_subroutines import *
from SLOPpy.subroutines.io_subroutines import *
from SLOPpy.subroutines.fit_subroutines import *
from SLOPpy.subroutines.shortcuts import *

__all__ = ['plot_compare_clv_rm_effects_planetRF',
           'plot_compare_clv_rm_effects_observerRF',
           'plot_compare_clv_rm_effects_stellarRF',
           'plot_compare_clv_rm_effects']


def plot_compare_clv_rm_effects_planetRF(config_in, night_input=''):
    plot_compare_clv_rm_effects(config_in, night_input, reference='planetRF')


def plot_compare_clv_rm_effects_observerRF(config_in, night_input=''):
    plot_compare_clv_rm_effects(config_in, night_input, reference='observerRF')


def plot_compare_clv_rm_effects_stellarRF(config_in, night_input=''):
    plot_compare_clv_rm_effects(config_in, night_input, reference='stellarRF')


def plot_compare_clv_rm_effects(config_in, night_input='', reference='planetRF'):
    
    transmission_average = load_from_cpickle('transmission_average_'+reference, config_in['output'])
    transmission_clv_rm_average = load_from_cpickle('transmission_clv_rm_average_'+reference, config_in['output'])
    night_dict = from_config_get_nights(config_in)

    fig = plt.figure(figsize=(12, 9))

    gs = GridSpec(2, 1)
    ax1 = plt.subplot(gs[0, 0])
    ax2 = plt.subplot(gs[1, 0], sharex=ax1, sharey=ax1)

    ax1.errorbar(transmission_clv_rm_average['wave'],
                 transmission_clv_rm_average['average'],
                 yerr=transmission_clv_rm_average['average_err'],
                 fmt='ko', ms=1, zorder=10, alpha=0.10, label='average, with CLV RM')

    ax1.errorbar(transmission_clv_rm_average['binned_wave'],
                 transmission_clv_rm_average['binned'],
                 yerr=transmission_clv_rm_average['binned_err'],
                 fmt='ko', ms=3, zorder=20, label='binned, with CLV RM')

    #ax1.errorbar(transmission_average['binned_wave'],
    #             transmission_average['binned'],
    #             yerr=transmission_average['binned_err'],
    #             fmt='mo', ms=3, zorder=15, label='binned, no CLV RM')


    ax2.errorbar(transmission_average['wave'],
                 transmission_average['average'],
                 yerr=transmission_average['average_err'],
                 fmt='ko', ms=1, zorder=10, alpha=0.10, label='average, no CLV RM')

    ax2.errorbar(transmission_average['binned_wave'],
                 transmission_average['binned'],
                 yerr=transmission_average['binned_err'],
                 fmt='ko', ms=3, zorder=20, label='binned, no CLV RM')

    #ax2.errorbar(transmission_clv_rm_average['binned_wave'],
    #             transmission_clv_rm_average['binned'],
    #             yerr=transmission_clv_rm_average['binned_err'],
    #             fmt='mo', ms=3, zorder=15, alpha=0.5, label='binned, with CLV RM')

    total_night = len(night_dict)
    side_step = config_in['master-out']['wavelength_step'] * config_in['master-out']['binning_factor'] / 10
    for n_night, night in enumerate(night_dict):
        ax1.errorbar(transmission_clv_rm_average['binned_wave'] + (n_night-total_night/2) * side_step,
                     transmission_clv_rm_average[night]['binned'],
                     yerr=transmission_clv_rm_average[night]['binned_err'],
                     color='C'+repr(n_night), label='{0:s} with CLV RM'.format(night),
                     fmt='o', ms=1, zorder=17, alpha=0.75)

        ax2.errorbar(transmission_average['binned_wave'] + (n_night-total_night/2) * side_step,
                     transmission_average[night]['binned'],
                     yerr=transmission_average[night]['binned_err'],
                     color='C' + repr(n_night), label='{0:s} no CLV RM'.format(night),
                     fmt='o', ms=1, zorder=17, alpha=0.75)


    #ax1.set_ylim(0.95-spec_offset*(1.+n_night), 1.05)
    ax1.set_xlim(config_in['master-out']['wavelength_range'][0], config_in['master-out']['wavelength_range'][1])
    ax1.set_ylim(0.985, 1.01)
    ax2.set_xlabel('$\lambda$ [$\AA$]')
    ax1.legend(loc=3)
    ax1.set_title('Average transmission spectrum with CLV and RM correction, in {0:s}'.format(reference))
    ax2.set_title('Average transmission spectrum without CLV and RM correction, in {0:s}'.format(reference))

    plt.show()




import numpy as np
from astropy.io import fits
from SLOPpy.subroutines.rebin_subroutines import *
from SLOPpy.subroutines import constants
from SLOPpy.subroutines.common import *

from astropy.coordinates import SkyCoord
from astropy import units as u

def PEPSI_get_instrument_keywords():

    properties = {
        # DRS-specific keywords
        'time_stamp': 'mid_exposure',
        'time_standard': 'TDB',

        # Observatory-specific keywords
        'geoelev': 3221.0, # meters
        'longitude' : -7.325938, # Tel geo longitude (+=East) (deg)
        'latitude' : 32.7013083,  # Tel geo latitute (+=North) (deg)

        # Instrument-specific keyword
        # The following are the input values used by Molecfit, taken from Allart+2017
        # for convenience, all the default values are listed here instead of being scattered into the code
        'molecfit': {
            'default_wstep': 0.01000, # default wavelength step size for the input stellar spectra
            'molecules': ['H2O', 'O2'],
            'ftol': 1e-9,
            'xtol': 1e-9,

            'cont_const': 1.0, # a0,  This value differs from Allart+2017 since we are using normalized spectra
            'cont_n': 3, # n_cont, Degree of coefficients for continuum fit
            'wlc_n': 2, # n_lambda, Polynomial degree of the refined wavelength solution
            'wlc_const': 0.0, # b0, Initial constant term for wavelength correction (shift relative to half  wavelength range)
            'res_gauss': 4.8, # omega_Gaussian,  Initial value for FWHM of Gaussian in pixels
            'kernfac': 15, #kernel_size, Size of Gaussian/Lorentzian/Voigtian kernel in FWHM

            'slitwidth': 1.00, # in arcseconds
            'pixelscale': 0.16,
        }
    }
    return properties

def PEPSI_get_calib_data(archive, file_rad, fiber='A', order_selection=None):
    """ There are no calibration files from PEPSI, so this subroutine will 
        provide the dictionary required by SLOPpy to properly work
        "fiber", "order_selection" variables are kept for consistency with the
        main code, but they do not have applicability 
    """
    calib_dict = {}

    # file from which to extract the relevant keywords - it could be a science
    # frame as well
    pepsi_fits = fits.open(archive+'/'+file_rad)
    data_fits = pepsi_fits[1].data['Fun']

    # Blaze file - if only blaze-corrected files are available, set it equal to 1.
    calib_dict['n_pixels'] = len(data_fits)
    calib_dict['n_orders'] = 1

    calib_dict['blaze'] = np.ones((calib_dict['n_orders'], calib_dict['n_pixels']))


    return calib_dict


def PEPSI_get_input_data(archive, file_rad, mask, fiber='A', skip_ccf=None, skip_s1d=True, order_selection=None):
    """_summary_

    Returns:
        _type_: _description_
    """

    """ PEPSI delivers calibrated, rebinned spectra only in a specific format
        so many entry in the dictionary will be empty.
        Given the simplicity of the format, this subroutines can be used as
        template for other instruments
        "fiber", "skip_ccf", "skip_s1d", "order_selection" variables are kept
        for consistency with the main code, but they do not have applicability
    """

    input_dict = {'mask': mask, 'header':{}}
    input_s1d = {'header':{}}

    properties = PEPSI_get_instrument_keywords()


    pepsi_fits = fits.open(archive+'/'+file_rad)
    input_dict['header']['e2ds'] = pepsi_fits[0].header


    #Arg = file_fits[1].data['Arg']
    #Fun = file_fits[1].data['Fun']
    #Var = file_fits[1].data['Var']
    #Mask = file_fits[1].data['Mask']

    input_dict['n_pixels'] = pepsi_fits[1].header['NAXIS2']
    input_dict['n_orders'] = 1

    # Not sure if these keywords are required anywhere
    #input_dict['DPR_CATG'] = 'SCIENCE'
    #input_dict['DPR_TYPE'] = ?? 

    input_dict['BERV'] = pepsi_fits[0].header['SSBVEL'] / 1000. # in km/s
    input_dict['RVC'] = pepsi_fits[0].header['RADVEL'] / 1000.
    # RV of the star, it must be provided but it can be bypassed


    input_dict['EXPTIME'] = pepsi_fits[0].header['EXPTIME']

    # BJD provided at midexposure ,no need to check for it
    input_dict['BJD'] = pepsi_fits[0].header['JD-TDB']
    input_dict['MJD'] = pepsi_fits[0].header['JD-OBS'] - constants.MJD

    input_dict['AIRMASS'] = pepsi_fits[0].header['AIRMASS']

    input_dict['UTC'] = (input_dict['MJD'] - int(input_dict['MJD'])) * 86400.
    input_dict['HUMIDITY'] = pepsi_fits[0].header['LBTH'] # Relative humidity in % for GEOELEV.
    input_dict['PRESSURE'] = pepsi_fits[0].header['LBTP']
    input_dict['TEMPERATURE_EN'] = pepsi_fits[0].header['LBTT']  #Ambient temperature in C for GEOELEV
    input_dict['TEMPERATURE_M1'] = pepsi_fits[0].header['LBTT']  #Temperature of primary mirror M1 in C (for emission spectra only)
    input_dict['ELEVATION'] = np.arcsin(1./input_dict['AIRMASS']) * (180./np.pi)

    input_dict['GEOELEV'] = properties['geoelev']
    input_dict['GEOLONG'] = properties['longitude']
    input_dict['GEOLAT'] = properties['latitude']

    input_dict['molecfit'] = properties['molecfit']

    skycoords = c = SkyCoord(pepsi_fits[0].header['RA'], pepsi_fits[0].header['DEC'], unit=(u.hourangle, u.deg))
    input_dict['RA'] = c.ra.degree
    input_dict['DEC'] = c.dec.degree

    # Not sure if thes evalies are required, it may be possible
    #input_dict['BLAZE_file'] = None
    #input_dict['CCD_SIGDET'] = None
    #input_dict['CCD_GAIN'] = None

    # getting data

    """
    NOTE: PEPSI provides 1D spectra in the stellar reference frame,
    but SLOPpy requires 2D spectra (order by order) in the observer reference
    frame. Thus:
    1) we shift the wavelength from the stellar to the observer reference frame
    2) we transform the array into (1,n_pixels) shaped arrays

    An empty array as required by SLOPpy would have the shape
    np.empty([n_orders, n_pixels])
    """


    wave_stellar =  pepsi_fits[1].data['Arg']
    rvshift = pepsi_fits[0].header['SSTVEL'] / 1000.

    input_dict['wave_size'] = pepsi_fits[1].header['NAXIS2']
    input_dict['wave'] = np.reshape(shift_wavelength_array(wave_stellar, rvshift), (1, input_dict['wave_size']))
    input_dict['e2ds'] = np.reshape(pepsi_fits[1].data['Fun'], (1, input_dict['wave_size']))
    input_dict['e2ds_err'] = np.reshape(pepsi_fits[1].data['Var'], (1, input_dict['wave_size']))

    """ PEPSI spectra are normalized to unity, but SLOPpy is expecting to have spectra in absolute counts
        absolute counts mean that a larger step size will have a larger number of counts given the same
        flux density at a specific wavelength.
        PEPSI spectra have been resampled on a non-linear scale and than normalized, so not taking into account
        the bin (or step) size would introduce a deformation during the rebinning phase
        After several tests, I am forced to introduce a flag to force non-preservation of flux at every
        rebinning step across the code

    """
    input_dict['absolute_flux'] = False

    input_dict['step'] = np.zeros_like(input_dict['wave'])
    input_dict['step'][0,1:-1] = (input_dict['wave'][0,2:] - input_dict['wave'][0,:-2])/2.
    input_dict['step'][0,0] = input_dict['step'][0,1]
    input_dict['step'][0,-1] = input_dict['step'][0,-2]

    # order selection is always equal the the first - and unique - order
    input_dict['orders'] = [0]

    input_dict['absolute_flux'] = False


    pepsi_fits.close()

    return input_dict,input_s1d



from __future__ import print_function, division
from SLOPpy.subroutines.common import *
from SLOPpy.subroutines.spectral_subroutines import *
from SLOPpy.subroutines.io_subroutines import *
from SLOPpy.subroutines.fit_subroutines import *
from SLOPpy.subroutines.plot_subroutines import *
from SLOPpy.subroutines.shortcuts import *

__all__ = ["check_differential_refraction", "plot_check_differential_refraction", "write_differential_refraction"]

subroutine_name = 'check_differential_refraction'


def check_differential_refraction(config_in):

    night_dict = from_config_get_nights(config_in)

    for night in night_dict:

        try:
            processed = load_from_cpickle('check_differential_refraction_processed', config_in['output'], night)
            check_drc = load_from_cpickle('check_differential_refraction', config_in['output'], night)
            print("{0:45s} Night:{1:15s}   {2:s}".format(subroutine_name, night, 'Retrieved'))
            continue
        except:
            print("{0:45s} Night:{1:15s}   {2:s}".format(subroutine_name, night, 'Computing'))
            print()

        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)

        calib_data = load_from_cpickle('calibration_fibA', config_in['output'], night)

        input_data = retrieve_observations(config_in['output'], night, lists['observations'],
                                           use_refraction=False, use_telluric=False)

        input_data_corrected = retrieve_observations(config_in['output'], night, lists['observations'],
                                           use_refraction=True, use_telluric=False)

        input_data_s1d = load_from_cpickle('input_dataset_s1d_fibA', config_in['output'], night)

        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

        check_drc = {
            'subroutine': subroutine_name,
            'wave': input_data['coadd']['wave']
        }
        processed = {
            'subroutine': subroutine_name
        }

        for obs in lists['observations']:

            processed[obs] = {
                'n_orders': input_data[obs]['n_orders'],
                'n_pixels': input_data[obs]['n_pixels']
            }

            """ for plotting purpose only"""
            #processed[obs]['e2ds'] = input_data[obs]['e2ds']
            #processed[obs]['e2ds_err'] = input_data[obs]['e2ds_err']

            #processed[obs]['flux'] = input_data[obs]['e2ds']/calib_data['blaze']/input_data[obs]['step']
            #processed[obs]['flux_err'] = np.sqrt(input_data[obs]['e2ds'])/calib_data['blaze']/input_data[obs]['step']
            preserve_flux = input_data[obs].get('absolute_flux', True)

            processed[obs]['flux_s1d'] = \
                rebin_2d_to_1d(input_data[obs]['wave'], input_data[obs]['step'], input_data[obs]['e2ds'],
                               calib_data['blaze'], input_data['coadd']['wave'], input_data['coadd']['step'],
                               preserve_flux=preserve_flux,
                               rv_shift=observational_pams[obs]['rv_shift_ORF2BRF'])

            """
            processed[obs]['flux_s1d_err'] = \
                rebin_2d_to_1d(input_data[obs]['wave'], input_data[obs]['step'],  input_data[obs]['e2ds_err'],
                               calib_data['blaze'], input_data['coadd']['wave'], input_data['coadd']['step'],
                               preserve_flux=preserve_flux,
                               rv_shift=0.00, is_error=True)
            """
            processed[obs]['flux_s1d_err'] = processed[obs]['flux_s1d']


            processed[obs]['s1d_rescaling'], processed[obs]['s1d_rescaled'], processed[obs]['s1d_rescaled_err'] = \
                perform_rescaling(input_data['coadd']['wave'],
                                  processed[obs]['flux_s1d'],
                                  processed[obs]['flux_s1d_err'],
                                  [5450.0, 5550.0])
                                  #observational_pams['wavelength_rescaling'])

            """ for plotting purpose only"""
            #processed[obs]['e2ds_corr'] = input_data_corrected[obs]['e2ds']
            #processed[obs]['e2ds_err_corr'] = input_data_corrected[obs]['e2ds_err']

            #processed[obs]['flux_corr'] = input_data_corrected[obs]['e2ds']/calib_data['blaze']/input_data_corrected[obs]['step']
            #processed[obs]['flux_err_corr'] = np.sqrt(input_data_corrected[obs]['e2ds'])/calib_data['blaze']/input_data_corrected[obs]['step']


            processed[obs]['flux_s1d_corr'] = \
                rebin_2d_to_1d(input_data_corrected[obs]['wave'],
                               input_data_corrected[obs]['step'],
                               input_data_corrected[obs]['e2ds'],
                               calib_data['blaze'],
                               input_data_corrected['coadd']['wave'],
                               input_data_corrected['coadd']['step'],
                               preserve_flux=preserve_flux,
                               rv_shift=observational_pams[obs]['rv_shift_ORF2BRF'])

            """  
            processed[obs]['flux_s1d_corr_err'] = \
                rebin_2d_to_1d(input_data_corrected[obs]['wave'],
                               input_data_corrected[obs]['step'],
                               input_data_corrected[obs]['e2ds_err'],
                               calib_data['blaze'],
                               input_data_corrected['coadd']['wave'],
                               input_data_corrected['coadd']['step'],
                               preserve_flux=preserve_flux,
                               rv_shift=0.00,
                               is_error=True)
            """
            processed[obs]['flux_s1d_corr_err'] = processed[obs]['flux_s1d_corr']

            processed[obs]['s1d_corr_rescaling'], processed[obs]['s1d_corr_rescaled'], processed[obs]['s1d_corr_rescaled_err'] = \
                perform_rescaling(input_data['coadd']['wave'],
                                  processed[obs]['flux_s1d_corr'],
                                  processed[obs]['flux_s1d_corr_err'],
                                  [5450.0, 5550.0])

            processed[obs]['dr_correction'] = processed[obs]['s1d_corr_rescaled']/processed[obs]['s1d_rescaled']

            processed[obs]['s1d_DRS_rescaling'], processed[obs]['s1d_DRS_rescaled'], processed[obs]['s1d_DRS_rescaled_err'] = \
                perform_rescaling(input_data_s1d[obs]['wave'],
                                  input_data_s1d[obs]['flux'],
                                  np.sqrt(np.abs(input_data_s1d[obs]['flux'])),
                                  [5450.0, 5550.0])
                                  #observational_pams['wavelength_rescaling'])

            processed[obs]['DRS_coeff_flux'] = []
            processed[obs]['DRS_corr'] = np.zeros(input_data_s1d[obs]['size'], dtype=np.double)

            for coeff_index in np.arange(0, 10, 1, dtype=np.int16):
                try:
                    keyword = 'HIERARCH TNG DRS FLUX CORR COEFF' + repr(coeff_index)
                    processed[obs]['DRS_coeff_flux'].extend([input_data[obs]['header']['ccf'][keyword]])
                    processed[obs]['DRS_corr'] += \
                        input_data[obs]['header']['ccf'][keyword] * \
                        np.power(input_data_s1d[obs]['wave'], coeff_index)
                except:
                    continue

            processed[obs]['DRS_corr_rescaling'], processed[obs]['DRS_corr_rescaled'], _ = \
                perform_rescaling(input_data_s1d[obs]['wave'],
                                  processed[obs]['DRS_corr'],
                                  processed[obs]['DRS_corr'],
                                  observational_pams['wavelength_rescaling'])

            check_drc[obs] = {
                's1d': {
                    'wave': input_data['coadd']['wave'],
                    'flux': processed[obs]['flux_s1d'],
                    #'flux_err': processed[obs]['flux_s1d_err'],
                    'rescaled': processed[obs]['s1d_rescaled'],
                    #'rescaled_err': processed[obs]['s1d_rescaled_err']
                    },
                's1d_corr': {
                    'correction': processed[obs]['dr_correction'],
                    'correction_rescaled': processed[obs]['dr_correction'],
                    'flux': processed[obs]['flux_s1d_corr'],
                    #'flux_err': processed[obs]['flux_s1d_corr_err'],
                    'rescaled': processed[obs]['s1d_corr_rescaled'],
                    #'rescaled_err': processed[obs]['s1d_corr_rescaled_err']
                    },
                'DRS_s1d':{
                    'wave': input_data_s1d[obs]['wave'],
                    'flux': input_data_s1d[obs]['flux'],
                    #'flux_err': np.sqrt(input_data_s1d[obs]['flux']+0.1),
                    'rescaled': processed[obs]['s1d_DRS_rescaled'],
                    #'rescaled_err': processed[obs]['s1d_DRS_rescaled_err']
                    },
                'DRS_s1d_corr': {
                    'correction': processed[obs]['DRS_corr'],
                    'correction_rescaled': processed[obs]['DRS_corr_rescaled'],
                    'flux': input_data_s1d[obs]['flux']/processed[obs]['DRS_corr'],
                    #'flux_err': np.sqrt(np.abs(input_data_s1d[obs]['flux']))/processed[obs]['DRS_corr'],
                    'rescaled': processed[obs]['s1d_DRS_rescaled']/processed[obs]['DRS_corr_rescaled'],
                    #'rescaled_err': processed[obs]['s1d_DRS_rescaled_err']/processed[obs]['DRS_corr_rescaled'],
                    },
            }

        save_to_cpickle('check_differential_refraction_processed', processed, config_in['output'], night)
        save_to_cpickle('check_differential_refraction', check_drc, config_in['output'], night)

        print('Night ', night, ' completed')
        print()


def plot_check_differential_refraction(config_in, night_input=''):
    night_dict = from_config_get_nights(config_in)

    if night_input == '':
        night_list = night_dict
    else:
        night_list = np.atleast_1d(night_input)

    for night in night_list:

        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)

        try:
            """ Retrieving the analysis"""
            check_drc = load_from_cpickle('check_differential_refraction', config_in['output'], night)
        except:
            print("  Failed in retrieving the data")
            return

        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

        """ Creation of the color array, based on the BJD of the observations
        """
        bjd = []
        am = []

        for obs in lists['observations']:
            bjd.append(observational_pams[obs]['BJD'] - 2450000.0)
            am.append(observational_pams[obs]['AIRMASS'])

        n_obs = len(lists['observations']) * 1.0 + 1.

        colors = np.asarray(bjd)
        cmap = plt.cm.viridis
        line_colors = cmap(np.linspace(0, 1, len(lists['observations'])))

        fig = plt.figure(figsize=(12, 6))
        gs = GridSpec(2, 2, width_ratios=[50, 1])
        gs.update(hspace=0.1)

        ax1 = plt.subplot(gs[0, 0])
        ax2 = plt.subplot(gs[1, 0], sharex=ax1, sharey=ax1)
        cbax1 = plt.subplot(gs[:, 1])

        for i, obs in enumerate(lists['observations']):

            sel = (check_drc[obs]['s1d']['rescaled'] > -1000000.05)

            ax1.plot(check_drc['wave'][sel],
                     check_drc[obs]['s1d']['rescaled'][sel]+i/5.,
                     c = line_colors[i], lw = 1, alpha = 1)

            ax2.plot(check_drc[obs]['DRS_s1d']['wave'],
                     check_drc[obs]['DRS_s1d']['rescaled']+i/5.,
                     c = line_colors[i], lw = 1, alpha = 1)

            i_max = 1.5 + i/5.

        ax1.set_xlim(check_drc['wave'][0], check_drc['wave'][-1])

        ax1.set_ylim(0.00, i_max)
        ax1.legend(loc=3)
        ax1.set_title('Night: {0:s} \n SLOPpy input s1d'.format(night))
        ax2.set_title('DRS input s1d')

        ax2.set_xlabel('$\lambda$ [$\AA$]')

        sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(vmin=colors[0], vmax=colors[-1]))
        sm.set_array([])  # You have to set a dummy-array for this to work...
        cbar = plt.colorbar(sm, cax=cbax1)
        cbar.set_label('BJD - 2450000.0')
        fig.subplots_adjust(wspace=0.05, hspace=0.4)
        plt.show()


        fig = plt.figure(figsize=(12, 6))
        gs = GridSpec(2, 2, width_ratios=[50, 1])
        #gs.update(wspace=0.025, hspace=0.05)
        gs.update(hspace=0.1)

        ax1 = plt.subplot(gs[0, 0])
        ax2 = plt.subplot(gs[1, 0], sharex=ax1, sharey=ax1)
        cbax1 = plt.subplot(gs[:, 1])

        for i, obs in enumerate(lists['observations']):

            sel = (check_drc[obs]['s1d']['flux']> 0.05)

            #ax1.plot(check_drc['wave'][sel],
            #         check_drc[obs]['s1d'][sel]+i/5.,
            #         c=line_colors[i], lw=1, alpha=1.0, zorder=0)

            ax1.plot(check_drc['wave'],
                     check_drc[obs]['s1d_corr']['correction_rescaled']+i/5.,
                     c=line_colors[i], lw=1, alpha=1)

            #ax2.plot(check_drc[obs]['wave_DRS'],
            #         check_drc[obs]['s1d_DRS']+i/5.,
            #         c=line_colors[i], lw=1, alpha=1)

            ax2.plot(check_drc[obs]['DRS_s1d']['wave'],
                     1./check_drc[obs]['DRS_s1d_corr']['correction_rescaled']+i/5.,
                     c=line_colors[i], lw=1, alpha=1)

            i_max = 1.5 + i/5.

        #ax1.plot(processed['coadd']['wave'], processed['coadd']['rescaled'], c='k', lw=1)
        #ax2.plot(processed['coadd']['wave'], processed['coadd']['rescaled'], c='k', lw=1)
        ax1.set_xlim(check_drc['wave'][0], check_drc['wave'][-1])

        ax1.set_ylim(0.00, i_max)
        ax1.legend(loc=3)
        ax1.set_title('Night: {0:s} \n SLOPpy correction function'.format(night))
        ax2.set_title('DRS correction function')

        ax1.set_xlabel('$\lambda$ [$\AA$]')

        sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(vmin=colors[0], vmax=colors[-1]))
        sm.set_array([])  # You have to set a dummy-array for this to work...
        cbar = plt.colorbar(sm, cax=cbax1)
        cbar.set_label('BJD - 2450000.0')
        fig.subplots_adjust(wspace=0.05, hspace=0.4)
        plt.show()

        fig = plt.figure(figsize=(12, 6))
        gs = GridSpec(2, 2, width_ratios=[50, 1])
        #gs.update(wspace=0.025, hspace=0.05)
        gs.update(hspace=0.1)

        ax1 = plt.subplot(gs[0, 0])
        ax2 = plt.subplot(gs[1, 0], sharex=ax1, sharey=ax1)
        cbax1 = plt.subplot(gs[:, 1])

        for i, obs in enumerate(lists['observations']):

            sel = (check_drc[obs]['s1d_corr']['rescaled']> 0.05)

            ax1.plot(check_drc['wave'][sel],
                     check_drc[obs]['s1d_corr']['rescaled'][sel]+i/5.,
                     c=line_colors[i], lw=1, alpha=0.5)

            ax2.plot(check_drc[obs]['DRS_s1d']['wave'],
                     check_drc[obs]['DRS_s1d_corr']['rescaled']+i/5.,
                     c=line_colors[i], lw=1, alpha=0.5)

            i_max = 1.5 + i/5.

        #ax1.plot(processed['coadd']['wave'], processed['coadd']['rescaled'], c='k', lw=1)
        #ax2.plot(processed['coadd']['wave'], processed['coadd']['rescaled'], c='k', lw=1)
        ax1.set_xlim(check_drc['wave'][0], check_drc['wave'][-1])

        ax1.set_ylim(0.00, i_max)
        ax1.legend(loc=3)
        ax1.set_title('Night: {0:s} \n SLOPpy corrected spectra'.format(night))
        ax2.set_title('DRS corrected spectra')
        ax2.set_xlabel('$\lambda$ [$\AA$]')

        sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(vmin=colors[0], vmax=colors[-1]))
        sm.set_array([])  # You have to set a dummy-array for this to work...
        cbar = plt.colorbar(sm, cax=cbax1)
        cbar.set_label('BJD - 2450000.0')
        fig.subplots_adjust(wspace=0.05, hspace=0.4)
        plt.show()


def write_differential_refraction(config_in):

    night_dict = from_config_get_nights(config_in)

    for night in night_dict:

        print()
        print('write_differential_refraction              Night: ', night)
        print()

        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)

        calib_data = load_from_cpickle('calibration_fibA', config_in['output'], night)

        input_data = retrieve_observations(config_in['output'], night, lists['observations'],
                                           use_refraction=True)

        input_data_s1d = load_from_cpickle('input_dataset_s1d_fibA', config_in['output'], night)

        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

        # Let's keep it simple to save memory

        dir_SLOPpy_drc = night + '_SLOPpy_drc/'
        dir_DRS_drc = night + '_DRS_drc/'

        os.system('mkdir -p ' + dir_SLOPpy_drc)
        os.system('mkdir -p ' + dir_DRS_drc)

        for obs in lists['observations']:

            processed = dict(n_orders=input_data[obs]['n_orders'], n_pixels=input_data[obs]['n_pixels'])

            preserve_flux = input_data[obs].get('absolute_flux', True)

            processed['flux_s1d_BRF_corr'] = \
                rebin_2d_to_1d(input_data[obs]['wave'], input_data[obs]['step'],input_data[obs]['e2ds'],
                               calib_data['blaze'], input_data['coadd']['wave'], input_data['coadd']['step'],
                               preserve_flux=preserve_flux,
                               rv_shift=observational_pams[obs]['rv_shift_ORF2BRF'])

            processed['flux_s1d_BRF_corr_err'] = \
                rebin_2d_to_1d(input_data[obs]['wave'], input_data[obs]['step'],input_data[obs]['e2ds_err'],
                               calib_data['blaze'], input_data['coadd']['wave'], input_data['coadd']['step'],
                               preserve_flux=preserve_flux,
                               rv_shift=observational_pams[obs]['rv_shift_ORF2BRF'], is_error=True)

            processed['flux_s1d_SRF_corr'] = \
                rebin_2d_to_1d(input_data[obs]['wave'], input_data[obs]['step'],input_data[obs]['e2ds'],
                               calib_data['blaze'], input_data['coadd']['wave'], input_data['coadd']['step'],
                               preserve_flux=preserve_flux,
                               rv_shift=observational_pams[obs]['rv_shift_ORF2SRF'])

            processed['flux_s1d_SRF_corr_err'] = \
                rebin_2d_to_1d(input_data[obs]['wave'], input_data[obs]['step'],input_data[obs]['e2ds_err'],
                               calib_data['blaze'], input_data['coadd']['wave'], input_data['coadd']['step'],
                               preserve_flux=preserve_flux,
                               rv_shift=observational_pams[obs]['rv_shift_ORF2SRF'], is_error=True)


            processed['DRS_coeff_flux'] = []
            processed['DRS_s1d_corr'] = np.zeros(input_data_s1d[obs]['size'], dtype=np.double)
            processed['DRS_e2ds_corr'] = np.zeros(np.shape(input_data[obs]['wave']), dtype=np.double)

            for coeff_index in np.arange(0, 10, 1, dtype=np.int16):
                try:
                    keyword = 'HIERARCH TNG DRS FLUX CORR COEFF' + repr(coeff_index)
                    processed['DRS_coeff_flux'].extend([input_data[obs]['header']['ccf'][keyword]])
                    processed['DRS_s1d_corr'] += \
                        input_data[obs]['header']['ccf'][keyword] * \
                        np.power(input_data_s1d[obs]['wave'], coeff_index)
                    processed['DRS_e2ds_corr'] += \
                        input_data[obs]['header']['ccf'][keyword] * \
                        np.power(input_data[obs]['wave'], coeff_index)
                except:
                    continue

            processed['DRS_s1d_corr'] = input_data_s1d[obs]['flux']/processed['DRS_s1d_corr']
            processed['DRS_e2ds_corr'] = input_data[obs]['e2ds']/processed['DRS_e2ds_corr']

            """Saving the e2ds files"""
            hdu_e2ds_SLOPpy = fits.PrimaryHDU()
            hdu_e2ds_DRS = fits.PrimaryHDU()

            hdu_e2ds_SLOPpy.data = np.asarray(input_data[obs]['e2ds'], dtype=np.float32)
            hdu_e2ds_DRS.data = np.asarray(processed['DRS_e2ds_corr'], dtype=np.float32)

            for key_name, key_val in input_data[obs]['header']['e2ds'].items():

                if key_name == 'SIMPLE' or \
                        key_name=='BITPIX' or \
                        key_name == 'NAXIS' or \
                        key_name == 'NAXIS1' or \
                        key_name == 'NAXIS2':
                    continue

                if len(key_name) > 8:
                    hdu_e2ds_SLOPpy.header['HIERARCH '+ key_name] = key_val
                    hdu_e2ds_DRS.header['HIERARCH '+ key_name] = key_val

                else:
                    hdu_e2ds_SLOPpy.header[key_name] = key_val
                    hdu_e2ds_DRS.header[key_name] = key_val

            hdu_e2ds_SLOPpy.writeto(dir_SLOPpy_drc + obs + '_e2ds_A.fits', overwrite=True)
            hdu_e2ds_DRS.writeto(dir_DRS_drc + obs + '_e2ds_A.fits', overwrite=True)

            """Saving the s1d files"""
            hdu_s1d_SLOPpy_BRF = fits.PrimaryHDU()
            hdu_s1d_SLOPpy_SRF = fits.PrimaryHDU()
            hdu_s1d_DRS = fits.PrimaryHDU()

            hdu_s1d_SLOPpy_BRF.data = np.asarray(processed['flux_s1d_BRF_corr'], dtype=np.float32)
            hdu_s1d_SLOPpy_SRF.data = np.asarray(processed['flux_s1d_SRF_corr'], dtype=np.float32)
            hdu_s1d_DRS.data = np.asarray(processed['DRS_s1d_corr'], dtype=np.float32)

            for key_name, key_val in input_data[obs]['header']['s1d'].items():

                if key_name == 'SIMPLE' or \
                        key_name=='BITPIX' or \
                        key_name == 'NAXIS' or \
                        key_name == 'NAXIS1' or \
                        key_name == 'NAXIS2':
                    continue

                if len(key_name) > 8:
                    hdu_s1d_SLOPpy_BRF.header['HIERARCH '+ key_name] = key_val
                    hdu_s1d_SLOPpy_SRF.header['HIERARCH '+ key_name] = key_val
                    hdu_s1d_DRS.header['HIERARCH '+ key_name] = key_val

                else:
                    hdu_s1d_SLOPpy_BRF.header[key_name] = key_val
                    hdu_s1d_SLOPpy_SRF.header[key_name] = key_val
                    hdu_s1d_DRS.header[key_name] = key_val

            """ Fixing SLOPpy s1d keywords """
            hdu_s1d_SLOPpy_BRF.header['CRVAL1'] = input_data['coadd']['wave'][0]
            hdu_s1d_SLOPpy_BRF.header['CDELT1'] = input_data['coadd']['step'][0]

            hdu_s1d_SLOPpy_SRF.header['CRVAL1'] = input_data['coadd']['wave'][0]
            hdu_s1d_SLOPpy_SRF.header['CDELT1'] = input_data['coadd']['step'][0]

            """ Fixing DRS s1d keywords """
            hdu_s1d_DRS.header['CRVAL1'] = input_data_s1d[obs]['wave'][0]
            hdu_s1d_DRS.header['CDELT1'] = input_data_s1d[obs]['step'][0]

            hdu_s1d_SLOPpy_BRF.writeto(dir_SLOPpy_drc + obs + '_s1d_A.fits', overwrite=True)
            hdu_s1d_SLOPpy_SRF.writeto(dir_SLOPpy_drc + obs + '_s1d_A_SRF.fits', overwrite=True)
            hdu_s1d_DRS.writeto(dir_DRS_drc + obs + '_s1d_A.fits', overwrite=True)

        print()
        print('Night ', night, ' completed')


from __future__ import print_function, division
from SLOPpy.subroutines.common import *
from SLOPpy.subroutines.spectral_subroutines import *
from SLOPpy.subroutines.io_subroutines import *
from SLOPpy.subroutines.fit_subroutines import *
from SLOPpy.subroutines.shortcuts import *
from SLOPpy.subroutines.rebin_subroutines import *
from SLOPpy.subroutines.clv_rm_subroutines import *

__all__ = ['write_output_spectra']

def write_output_spectra(config_in):

    subroutine_name = 'write_output_spectra'
    clv_rm_correction = True

    night_dict = from_config_get_nights(config_in)
    instrument_dict = from_config_get_instrument(config_in)
    system_dict = from_config_get_system(config_in)
    planet_dict = from_config_get_planet(config_in)

    shared_data = load_from_cpickle('shared', config_in['output'])

    lightcurve_dict = from_config_get_transmission_lightcurve(config_in)

    for night in night_dict:

        clv_rm_correction = True
        try:
            clv_rm_modelling = load_from_cpickle('clv_rm_modelling', config_in['output'], night)
        except:
            clv_rm_correction = False

        message = 'Computing'
        try:
            output_spectra = load_from_cpickle(subroutine_name, config_in['output'], night)

            if clv_rm_correction and not output_spectra['clv_rm_correction']:
                message = 'Updating with CLV-corrected spectra'
                raise ValueError()

            print("{0:45s} Night:{1:15s}   {2:s}".format(subroutine_name, night, 'Retrieved'))
            continue
        except:
            print("{0:45s} Night:{1:15s}   {2:s}".format(subroutine_name, night, message))
            print()

        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)

        calib_data = load_from_cpickle('calibration_fibA', config_in['output'], night)
        input_data = retrieve_observations( config_in['output'], night, lists['observations'])
        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

        output_spectra = {
            'subroutine': subroutine_name,
            'clv_rm_correction': clv_rm_correction
        }
        """ Adding the C-bands arrays to the dictionary"""

        if clv_rm_correction:
            clv_rm_modelling = load_from_cpickle('clv_rm_modelling', config_in['output'], night)

        for n_obs, obs in enumerate( lists['observations']):

            output_spectra[obs] = {}

            output_spectra[obs]['BJD'] = input_data[obs]['BJD']
            preserve_flux = input_data[obs].get('absolute_flux', True)

            output_spectra[obs]['SRF_rebinned'] = \
                rebin_2d_to_1d(input_data[obs]['wave'],
                               input_data[obs]['step'],
                               input_data[obs]['e2ds'],
                               calib_data['blaze'],
                               shared_data['coadd']['wave'],
                               shared_data['coadd']['step'],
                               preserve_flux=preserve_flux,
                               rv_shift=observational_pams[obs]['rv_shift_ORF2SRF'])

            output_spectra[obs]['SRF_rebinned_err'] = \
                rebin_2d_to_1d(input_data[obs]['wave'],
                               input_data[obs]['step'],
                               input_data[obs]['e2ds_err'],
                               calib_data['blaze'],
                               shared_data['coadd']['wave'],
                               shared_data['coadd']['step'],
                               preserve_flux=preserve_flux,
                               rv_shift=observational_pams[obs]['rv_shift_ORF2SRF'])

            output_spectra[obs]['SRF_rescaling'], \
            output_spectra[obs]['SRF_rescaled'], \
            output_spectra[obs]['SRF_rescaled_err'] = perform_rescaling(
                shared_data['coadd']['wave'], output_spectra[obs]['SRF_rebinned'], output_spectra[obs]['SRF_rebinned_err'],
                observational_pams['wavelength_rescaling'])

            if clv_rm_correction:

                rv_shift = 0.0 # we always stay in SRF
                correction, _ = clv_rm_correction_factor_computation(
                    clv_rm_modelling, shared_data['coadd']['wave'], shared_data['coadd']['step'], rv_shift, obs)

                output_spectra[obs]['SRF_clv_rm_correction'] = correction
                output_spectra[obs]['SRF_clv_rm_rebinned'] = output_spectra[obs]['SRF_rebinned'] / correction
                output_spectra[obs]['SRF_clv_rm_rebinned_err'] = output_spectra[obs]['SRF_rebinned_err'] / correction

                output_spectra[obs]['SRF_clv_rm_rescaled'] = output_spectra[obs]['SRF_rescaled'] / correction
                output_spectra[obs]['SRF_clv_rm_rescaled_err'] = output_spectra[obs]['SRF_rescaled_err'] / correction

            try:
                output_spectra[obs]['phase'] = \
                    (observational_pams[obs]['BJD'] - night_dict[night]['time_of_transit'][0])/planet_dict['period'][0]
            except:
                output_spectra[obs]['phase'] = \
                    (observational_pams[obs]['BJD'] - night_dict[night]['time_of_transit'])/planet_dict['period'][0]

        save_to_cpickle(subroutine_name, output_spectra, config_in['output'], night)




from __future__ import print_function, division
from SLOPpy.subroutines.io_subroutines import *
from SLOPpy.subroutines.common import *

def retrieve_observations(output_name, night, observations,
                          use_refraction=True, use_telluric=True, use_interstellar=True,
                          use_telluric_spline= False):

    #   config_in['output'], night, lists['observations']

    """ Retrieving the observations"""
    try:
        sky_corrected = load_from_cpickle('skycorrected_fibA', output_name, night)
        input_data = load_from_cpickle('input_dataset_fibA', output_name, night)
        for obs in observations:
            input_data[obs]['e2ds'] = sky_corrected[obs]['e2ds'].copy()
        print("  Sky contamination correction: applied ")
    except:
        input_data = load_from_cpickle('input_dataset_fibA', output_name, night)

    try:
        """ Retrieving the telluric correction"""
        telluric = load_from_cpickle('telluric', output_name, night)
        correct_telluric = True
    except:
        correct_telluric = False
        if use_telluric:
            print("  No telluric correction available - is this expected? ")

    """ Retrieval of refraction correction, if present """
    try:
        try:
            refraction = load_from_cpickle('refraction_update', output_name, night)
            correct_refraction = True
        except:
            refraction = load_from_cpickle('refraction', output_name, night)
            correct_refraction = True
    except:
        correct_refraction = False
        if use_refraction:
            print("  No refraction correction available - is this expected? ")

    """ Retrieval of refraction correction, if present """
    try:
        interstellar = load_from_cpickle('interstellar_lines', output_name, night)
        correct_interstellar = True
    except:
        correct_interstellar = False
        if use_interstellar:
            print("  No interstellar lines correction available - is this expected?  ")

    for obs in observations:

        """ Spectra are collected for variations of differential refraction,
        if the correction has been computed  before"""
        if correct_refraction and use_refraction:
            try:
                input_data[obs]['e2ds'] /= refraction[obs]['fit_e2ds']
                input_data[obs]['e2ds_err'] /= refraction[obs]['fit_e2ds']
            except:
                input_data[obs]['e2ds'] /= refraction[obs]['polyfit_e2ds']
                input_data[obs]['e2ds_err'] /= refraction[obs]['polyfit_e2ds']

        if correct_telluric and use_telluric:
            if use_telluric_spline:
                input_data[obs]['e2ds'] /= telluric[obs]['spline']
                input_data[obs]['e2ds_err'] /= telluric[obs]['spline']
            else:
                input_data[obs]['e2ds'] /= telluric[obs]['spectrum']
                input_data[obs]['e2ds_err'] /= telluric[obs]['spectrum']

        if correct_interstellar and use_interstellar:
            input_data[obs]['e2ds'] /= interstellar[obs]['correction']
            input_data[obs]['e2ds_err'] /= interstellar[obs]['correction']


    if correct_refraction and use_refraction:
        print("  Differential refraction correction: applied ")
    if correct_telluric and use_telluric:
        print("  Telluric correction: applied ")
    if correct_interstellar and use_interstellar:
        print("  Interstellar lines correction: applied ")

    print()
    return input_data


def compute_rescaling(wave, flux, wavelength_range):

    wl_med = np.average(wavelength_range)
    wl_dif = wavelength_range[1]-wl_med
    sel = np.where(np.abs(wave-wl_med) < wl_dif)
    return np.median(flux[sel])


def perform_rescaling(wave, flux, ferr, wavelength_range):

    factor = compute_rescaling(wave, flux, wavelength_range)
    # flux_out = flux / factor
    # ferr_out = ferr / factor

    return factor, flux / factor, ferr / factor


def replace_values(vals, threshold, replacement=None):
    if not replacement:
        replacement = np.median(vals)

    null = (vals <= threshold)
    vals[null] = replacement
    return vals, null


def replace_values_errors(vals, errs, threshold, replacement=None, error_replacement=None):
    if not replacement:
        replacement = np.median(vals)
    if not error_replacement:
        error_replacement = np.median(errs)*10.0

    null = (vals <= threshold)
    vals[null] = replacement
    errs[null] = error_replacement
    return vals, errs, null


def replace_values_errors_with_interpolation_1d(vals,
                                                errs=None,
                                                less_than=None,
                                                greater_than=None,
                                                force_positive=False,
                                                error_replacement=None,
                                                sigma_iter=1):
    """
    Replace outliers in the input array and the associated arrors array (if provided)
    Written as a fix to out-of-boundaries rebinning
    At least one between  less_than  or  greater_than  must be provided

    :param vals: input array to be fixed
    :param errs: error array associated to input array
    :param less_than: lower threshold for considering a value as outlier.
    :param greater_than: upper threshold for considering a value as outlier.
    :param error_replacement: optional, error to be associated to the outlier.
        if not provided, ten times the error of the interpolated points is taken
    :param sigma_iter: optional, if provided it will consider less_than/greater_than as sigma multiplicators
    :return: vals, errs, null
    """

    if errs is None:
        errt = vals*0.0
    else:
        errt = errs

    if sigma_iter <= 1:
        use_sigma = False
    else:
        use_sigma = True

    if force_positive:
        null = (vals <= 0.00000000001)
    else:
        null = (~np.isfinite(vals))

    for ii in range(0, sigma_iter):

        if use_sigma:
            median = np.median(vals)
            sigma = np.std(vals)
            if less_than:
                threshold_less = median - sigma*less_than
            if greater_than:
                threshold_more = median + sigma*greater_than
        else:
            if less_than:
                threshold_less = less_than
            if greater_than:
                threshold_more = greater_than

        if less_than and greater_than:
            ind_sel = np.where((vals < threshold_less) | (vals > threshold_more) | null)[0]
        elif less_than:
            ind_sel = np.where((vals < threshold_less) | null)[0]
        elif greater_than:
            ind_sel = np.where((vals > threshold_more) | null)[0]
        elif force_positive:
            ind_sel = np.where((vals < 0.00001) | null)[0]
        else:
            raise ValueError('Provide at least on condition')


        null[ind_sel] = True
        i_e = 0
        for jj in ind_sel:
            if jj < i_e: continue
            i_s = jj
            i_e = jj

            while i_s in ind_sel:
                i_s -= 1
                if i_s < 0:
                    break

            while i_e in ind_sel:
                i_e += 1
                if i_e >= vals.size:
                    break

            if i_s < 0 and i_e >= vals.size:
                raise ValueError('Invalid input, something must be wrong in the input array')
            elif i_s < 0:
                vals[:i_e] = vals[i_e]
                errt[:i_e] = errt[i_e]*10.
            elif i_e >= vals.size:
                vals[i_s:] = vals[i_s]
                errt[i_s:] = errt[i_s]*10.
            else:
                vals[i_s+1:i_e] = (vals[i_e]+vals[i_s])/2.
                errt[i_s+1:i_e] = (vals[i_e]+vals[i_s])*10.

    if errs is None:
        return vals, null
    else:
        if error_replacement:
            errs[ind_sel] = error_replacement
        else:
            errs[ind_sel] = errt[ind_sel]
        return vals, errs, null


def replace_values_errors_with_interpolation_2d(vals,
                                                errs=None,
                                                less_than=None,
                                                greater_than=None,
                                                force_positive=False,
                                                error_replacement=None,
                                                sigma_iter=1,
                                                axis=0):

    if errs is None:
        errt = vals*0.0
    else:
        errt = errs

    null = (~np.isfinite(vals))

    if axis ==0:
        for n_order in range(0, np.size(vals[:,0])):
            vals[n_order, :], errt[n_order,:], null[n_order, :] = replace_values_errors_with_interpolation_1d(
                vals[n_order, :],
                errt[n_order, :],
                less_than,
                greater_than,
                force_positive,
                error_replacement,
                sigma_iter)
    else:
        for n_order in range(0, np.size(vals[0,:])):
            vals[:, n_order], errt[:, n_order], null[:, n_order] = replace_values_errors_with_interpolation_1d(
                vals[:, n_order],
                errt[:, n_order],
                less_than,
                greater_than,
                force_positive,
                error_replacement,
                sigma_iter)

    if errs is None:
        return vals, null
    else:
        return vals, errs, null


def compute_spline(wave_in, flux_in, knots_step, knot_order=3, use_null=None, use_selection=None):
    """ Sometimes the interpolation range is just outside the spline evaluation range. The spline procedure will
    fail even if the affected point are at the very extremes of the spectral range, i.e., points that would not be
    inin the analysis anyway. We trick the spline procedure by adding fake points
    """

    if use_null is not None:
        selection = (~use_null)
    elif use_selection is not None:
        selection = use_selection
    else:
        selection = (wave_in > 0.000)

    len_wave = np.sum(selection) + 6
    wave = np.empty(len_wave, dtype=np.double)
    spec = np.empty(len_wave, dtype=np.double)

    wave[:3] = wave_in[0] - np.linspace(1.0, 0.5, 3)
    wave[3:-3] = wave_in[selection]
    wave[-3:] = wave_in[-1] + np.linspace(0.5, 1.0, 3)

    spec[:3] = flux_in[selection][0]
    spec[3:-3] = flux_in[selection]
    spec[-3:] = flux_in[selection][-1]

    """ computing the spline approximation of the rescaled master out"""

    """ picking the number of knots """
    nknots = (np.amax(wave) - np.amin(wave)) /knots_step
    """ picking the indices of the knots"""
    idx_knots = (np.arange(1, len(wave) - 1, (len(wave) - 2.) / nknots)).astype('int')

    """ passing from indices to knots values """
    spline_knots = wave[idx_knots]

    spline_coeff = sci_int.splrep(wave, spec, task=-1, k=knot_order, t=spline_knots)
    spline_eval = sci_int.splev(wave_in, spline_coeff)

    return spline_eval, spline_coeff, spline_knots


def f_telluric_rescaling_factor(x, x_range, tel1, tel2):
    return np.sum((tel1-(1.+x[0]*x_range)-(tel2-1.)*x[1])**2)


def find_telluric_rescaling_factor(tel1, tel2):
    fit_selection = (tel2 < 1.0)

    x_range = np.arange(0.0, len(tel1), 1.0)/len(tel1)
    x_start = np.zeros(2)
    x_start[1] = np.median((tel1[fit_selection]-1.)/(tel2[fit_selection]-1.))
    x_result = sci_optimize.minimize(f_telluric_rescaling_factor,
                                     x_start,
                                     args=(x_range, tel1, tel2),
                                     method='Nelder-Mead',
                                     options=dict(maxiter=10000))
    return x_result['x'][1], x_result['x'][0], x_result['success']



def f_telluric_xslope(x, x_range, tel1, tel2, std):
    return np.sum((tel1-(1.+x*x_range))**2/std**2)


def f_telluric_factor(x, tel1, tel2, std):
    print(x, np.sum(((tel1-1.)-(tel2-1.)*x)**2/std**2))
    return np.sum(((tel1-1.)-(tel2-1.)*x)**2/std**2)


def find_telluric_rescaling_factor_2steps(tel1, tel2):
    fit_xslope = (tel2 > 0.99999)
    fit_factor = (tel2 < 0.99999)
    std_all = np.std(tel1)

    x_range = np.arange(0.0, len(tel1), 1.0)/len(tel1)
    x_start = 0.000

    x_result = sci_optimize.minimize(f_telluric_xslope,
                                     x_start,
                                     args=(x_range[fit_xslope], tel1[fit_xslope], tel2[fit_xslope], std_all),
                                     method='Nelder-Mead',
                                     options=dict(maxiter=10000))

    x_xslope = x_result['x'][0]
    x_start = np.median((tel1[fit_factor]-1.-x_range[fit_factor]*x_xslope)/(tel2[fit_factor]-1.))
    print('---> ', x_start)
    x_result = sci_optimize.minimize(f_telluric_factor,
                                     x_start,
                                     args=(tel1[fit_factor]-x_range[fit_factor]*x_xslope, tel2[fit_factor], std_all),
                                     method='Nelder-Mead',
                                     options=dict(maxiter=10000))

    x_factor = x_result['x'][0]

    return x_factor, x_xslope, x_result['success']




def write_molecfit_v1_par(filename_par, filename_data,  filename_output, filename_include, molecfit_dict, observing_dict):


    """ The observing_dict must contains these keywords:
    'MJD': Observing date in years or MJD in days
    'UTC': UTC in s
    'ELEVATION': Telescope altitude angle in deg
    'HUMIDITY': Humidity in %
    'PRESSURE': Pressure in hPa
    'TEMPERATURE_EN': Ambient temperature in deg C
    'TEMPERATURE_M1': Mirror temperature in deg C
    'GEOELEV': Elevation above sea level in m (default is Paranal: 2635m)
    'GEOLONG': Longitude
    'GEOLAT': Latitude
    """

    fileout = open(filename_par, 'w')
    fileout.write("### Driver for MOLECFIT\n")

    # user working directory only important for REFLEX workflow and GUI
    # not used by molecfit itself.
    fileout.write("user_workdir:./\n")

    ## INPUT DATA
    # Data file name (path relative to the current directory or absolute path)
    fileout.write("filename: " + filename_data + "\n")

    # ASCII list of files to be corrected for telluric absorption using the
    # transmission curve derived from the input reference file (path of list and
    # listed files relative to the current directory or absolute path; default: "none")
    fileout.write("listname: none\n")

    # Type of input spectrum -- 1 = transmission (default); 0 = emission
    fileout.write("trans: 1\n")

    # Names of the file columns (table) or extensions (image) containing:
    # Wavelength  Flux  Flux_Err  Mask
    # - Flux_Err and/or Mask can be avoided by writing 'NULL'
    # - 'NULL' is required for Wavelength if it is given by header keywords
    # - parameter list: col_lam, col_flux, col_dflux, and col_mask
    fileout.write("columns: Wavelength Flux NULL NULL\n")

    # Default error relative to mean for the case that the error column is missing
    fileout.write("default_error: 0.001\n")

    # Multiplicative factor to convert wavelength to micron
    # (e.g. nm -> wlgtomicron = 1e-3)
    fileout.write("wlgtomicron: 0.0001\n")

    # Wavelengths in vacuum (= vac) or air (= air)
    fileout.write("vac_air: air\n")

    # TODO: input from configuration file for molecfit installation path
    # ASCII or FITS table for wavelength ranges in micron to be fitted
    # (path relative to the current directory or absolute path; default: "none")
    fileout.write("wrange_include: " + filename_include + "\n")

    # ASCII or FITS table for wavelength ranges in micron to be excluded from the
    # fit (path relative to the current directory or absolute path; default: "none")
    # wrange_exclude: /Users/malavolta/Astro/ExoAtmospheres/molecfit_test//HIP63901_exclude_w.dat

    # ASCII or FITS table for pixel ranges to be excluded from the fit
    # (path relative to the current directory or absolute path; default: "none")
    # prange_exclude: /Users/malavolta/Astro/ExoAtmospheres/molecfit_test//HIP63901_exclude_p.dat

    ## RESULTS

    # Directory for output files (path relative to the current directory or absolute path)
    fileout.write("output_dir:./\n")

    # Name for output files
    # (supplemented by "_fit" or "_tac" as well as ".asc", ".atm", ".fits",
    # ".par, ".ps", and ".res")
    fileout.write("output_name: "+ filename_output + "\n")

    # Plot creation: gnuplot is used to create control plots
    # W - screen output only (incorporating wxt terminal in gnuplot)
    # X - screen output only (incorporating x11 terminal in gnuplot)
    # P - postscript file labelled '<output_name>.ps', stored in <output_dir>
    # combinations possible, i.e. WP, WX, XP, WXP (however, keep the order!)
    # all other input: no plot creation is performed
    fileout.write("plot_creation: none\n")

    # Create plots for individual fit ranges? -- 1 = yes; 0 = no
    fileout.write("plot_range: 0\n")

    ## FIT PRECISION

    # --FTOL
    # Relative chi-square convergence criterion.
    # FTOL=1e-9
    try:
        fileout.write('ftol: {0:.12f}\n'.format(molecfit_dict['ftol']))
    except ValueError:
        fileout.write("ftol=" + molecfit_dict['ftol'] + "\n")

    # --XTOL
    # Relative parameter convergence criterion.
    # XTOL=1e-10
    try:
        fileout.write('xtol: {0:.12f}\n'.format(molecfit_dict['xtol']))
    # Relative chi2 convergence criterion
    except ValueError:
        fileout.write("xtol: " + molecfit_dict['xtol'] + "\n")

    ## MOLECULAR COLUMNS

    # List of molecules to be included in the model
    # (default: 'H2O', N_val: nmolec)
    molecules_list = "list_molec:"
    for mol in molecfit_dict['molecules']:
        molecules_list += " " + mol
    fileout.write(molecules_list + "\n")

    # Fit flags for molecules -- 1 = yes; 0 = no (N_val: nmolec)
    fileout.write("fit_molec: 1 1\n")

    # Values of molecular columns, expressed relatively to the input ATM profile
    # columns (N_val: nmolec) [1 = 100%]
    fileout.write("relcol: 1.0 1.0\n")

    ## BACKGROUND AND CONTINUUM

    # Conversion of fluxes from phot/(s*m2*mum*as2) (emission spectrum only) to
    # flux unit of observed spectrum:
    # 0: phot/(s*m^2*mum*as^2) [no conversion]
    # 1: W/(m^2*mum*as^2)
    # 2: erg/(s*cm^2*A*as^2)
    # 3: mJy/as^2
    # For other units the conversion factor has to be considered as constant term
    # of the continuum fit.
    fileout.write("flux_unit: 0\n")

    # Fit of telescope background -- 1 = yes; 0 = no (emission spectrum only)
    fileout.write("fit_back: 0\n")

    # Initial value for telescope background fit (range: [0,1])
    fileout.write("telback: 0.1\n")

    # Polynomial fit of continuum --> degree: cont_n
    fileout.write("fit_cont: 1\n")

    # Degree of coefficients for continuum fit
    fileout.write("cont_n: {0:1.0f}".format(molecfit_dict['cont_n']) + "\n")

    # Initial constant term for continuum fit (valid for all fit ranges)
    # (emission spectrum: about 1 for correct flux_unit)
    fileout.write("cont_const: {0:1.0f}".format(molecfit_dict['cont_const']) + "\n")

    ## WAVELENGTH SOLUTION

    # Refinement of wavelength solution using a polynomial of degree wlc_n
    fileout.write("fit_wlc: 1\n")

    # Polynomial degree of the refined wavelength solution
    fileout.write("wlc_n: {0:1.0f}".format(molecfit_dict['wlc_n']) + "\n")

    # Initial constant term for wavelength correction (shift relative to half
    # wavelength range)
    fileout.write("wlc_const: {0:1.0f}".format(molecfit_dict['wlc_const']) + "\n")

    ## RESOLUTION

    # Fit resolution by boxcar -- 1 = yes; 0 = no
    fileout.write("fit_res_box: 0\n")

    # Initial value for FWHM of boxcar relative to slit width (>= 0. and <= 2.)
    fileout.write("relres_box: 0.0\n")

    # Voigt profile approximation instead of independent Gaussian and Lorentzian
    # kernels? -- 1 = yes; 0 = no
    fileout.write("kernmode: 0\n")

    # Fit resolution by Gaussian -- 1 = yes; 0 = no
    fileout.write("fit_res_gauss: 1\n")

    # Initial value for FWHM of Gaussian in pixels
    fileout.write("res_gauss: {0:3.1f}".format(molecfit_dict['res_gauss']) + "\n")

    # Fit resolution by Lorentzian -- 1 = yes; 0 = no
    fileout.write("fit_res_lorentz: 0\n")

    # Initial value for FWHM of Lorentzian in pixels
    fileout.write("res_lorentz: 0.0\n")

    # Size of Gaussian/Lorentzian/Voigtian kernel in FWHM
    fileout.write("kernfac: {0:3.0f}".format(molecfit_dict['kernfac']) + "\n")

    # Variable kernel (linear increase with wavelength)? -- 1 = yes; 0 = no
    fileout.write("varkern: 0\n")

    # ASCII file for kernel elements (one per line; normalisation not required)
    # instead of synthetic kernel consisting of boxcar, Gaussian, and Lorentzian
    # components (path relative to the current directory or absolute path; default: "none\n")
    fileout.write("kernel_file: none\n")

    ## AMBIENT PARAMETERS

    # If the input data file contains a suitable FITS header, the keyword names of
    # the following parameters will be read, but the corresponding values will not
    # be used. The reading of parameter values from this file can be forced by
    # setting keywords to NONE.

    # Observing date in years or MJD in days
    # changed the format to inter in order to not have molecfit crashed
    #fileout.write("obsdate: {0:13.5f}".format(observing_dict['MJD']) + "\n")
    fileout.write("obsdate: {0:13.50}".format(observing_dict['MJD']) + "\n")
    fileout.write("obsdate_key: NONE\n")

    # UTC in s
    fileout.write("utc: {0:8.0f}".format(observing_dict['UTC']) + "\n")
    fileout.write("utc_key: NONE\n")

    # Telescope altitude angle in deg
    fileout.write("telalt: {0:13.5f}".format(observing_dict['ELEVATION']) + "\n")
    fileout.write("telalt_key: NONE\n")

    # Humidity in %
    fileout.write("rhum: {0:13.5f}".format(observing_dict['HUMIDITY']) + "\n")
    fileout.write("rhum_key: NONE\n")

    # Pressure in hPa
    fileout.write("pres: {0:5.1f}".format(observing_dict['PRESSURE']) + "\n")
    fileout.write("pres_key: NONE\n")

    # Ambient temperature in deg C
    fileout.write("temp: {0:4.1f}".format(observing_dict['TEMPERATURE_EN']) + "\n")
    fileout.write("temp_key: NONE\n")

    # Mirror temperature in deg C
    fileout.write("m1temp: {0:4.1f}".format(observing_dict['TEMPERATURE_M1']) + "\n")
    fileout.write("m1temp_key: NONE\n")

    # Elevation above sea level in m (default is Paranal: 2635m)
    fileout.write("geoelev: {0:4.0f}".format(observing_dict['GEOELEV']) + "\n")
    fileout.write("geoelev_key: NONE\n")

    # Longitude (default is Paranal: -70.4051)
    fileout.write("longitude: {0:9.4f}".format(observing_dict['GEOLONG']) + "\n")
    fileout.write("longitude_key: NONE\n")

    # Latitude (default is Paranal: -24.6276)
    fileout.write("latitude: {0:9.4f}".format(observing_dict['GEOLAT']) + "\n")
    fileout.write("latitude_key: NONE\n")

    ## INSTRUMENTAL PARAMETERS

    # Slit width in arcsec (taken from FITS header if present)
    fileout.write("slitw: {0:3.1f}".format(molecfit_dict['slitwidth']) + "\n")
    fileout.write("slitw_key: NONE\n")

    # Pixel scale in arcsec (taken from this file only)
    fileout.write("pixsc: {0:4.2f}".format(molecfit_dict["pixelscale"]) + "\n")
    fileout.write("pixsc_key: NONE\n")

    ## ATMOSPHERIC PROFILES

    # Reference atmospheric profile
    fileout.write("ref_atm: equ.atm\n")

    # Specific GDAS-like input profile (P[hPa] HGT[m] T[K] RELHUM[%]) (path
    # relative to the installation directory or absolute path). In the case of "none", no GDAS
    # profiles will be considered. The default "auto" performs an automatic
    # retrieval.
    fileout.write("gdas_dir: data/profiles/grib\n")
    fileout.write("gdas_prof: auto\n")

    # Grid of layer heights for merging ref_atm and GDAS profile. Fixed grid = 1
    # (default) and natural grid = 0.
    fileout.write("layers: 0\n")

    # Upper mixing height in km (default: 5) for considering data of a local meteo
    # station. If emix is below geoelev, rhum, pres, and temp are not used for
    # modifying the corresponding profiles.
    fileout.write("emix: 5.0\n")

    # PWV value in mm for the input water vapour profile. The merged profile
    # composed of ref_atm, GDAS, and local meteo data will be scaled to this value
    # if pwv > 0 (default: -1 -> no scaling).
    fileout.write("pwv: -1.\n")

    # internal GUI specific parameter
    fileout.write("clean_mflux: 1\n")

    fileout.write("end\n")
    fileout.close()



def write_molecfit_par(filename_par, wave_include, molecfit_dict, observing_dict):


    """ The observing_dict must contains these keywords:
    'MJD': Observing date in years or MJD in days
    'UTC': UTC in s
    'ELEVATION': Telescope altitude angle in deg
    'HUMIDITY': Humidity in %
    'PRESSURE': Pressure in hPa
    'TEMPERATURE_EN': Ambient temperature in deg C
    'TEMPERATURE_M1': Mirror temperature in deg C
    'GEOELEV': Elevation above sea level in m (default is Paranal: 2635m)
    'GEOLONG': Longitude
    'GEOLAT': Latitude
    """

    fileout = open(filename_par, 'w')

    # File: molecfitConfigFiles/model.rc
    #
    # Note: This configuration file has been automatically
    #       generated by the esorex (v3.13.5) program.
    #
    # Date: 25-Jun-2022 16:52:38
    #
    #

    # --USE_ONLY_INPUT_PRIMARY_DATA
    # Value=TRUE implies that only the fits primary contains the input science flux
    # data.
    # Value=FALSE implies that the fits extensions also contains input science
    # flux data.
    fileout.write('USE_ONLY_INPUT_PRIMARY_DATA=FALSE\n')

    # --USE_DATA_EXTENSION_AS_DFLUX
    # Only valid if USE_ONLY_INPUT_PRIMARY_DATA=TRUE. The fits extension index that
    # contains the
    # errors of the science flux data (DFLUX). A value of 0 implies that there is
    # no DFLUX.
    fileout.write('USE_DATA_EXTENSION_AS_DFLUX=0\n')

    # --USE_DATA_EXTENSION_AS_MASK
    # Only valid if USE_ONLY_INPUT_PRIMARY_DATA=TRUE. The fits extension index that
    # contains the
    # mask associated with the science flux data. A value of 0 implies that there
    # is no mask data.
    fileout.write('USE_DATA_EXTENSION_AS_MASK=0\n')

    # --USE_INPUT_KERNEL
    # If TRUE, use the kernel library if it is provided.
    fileout.write('USE_INPUT_KERNEL=TRUE\n')

    # --MODEL_MAPPING_KERNEL
    # Mapping 'STD_MODEL/SCIENCE' - 'MODEL_KERNEL_LIBRARY' [string with ext_number
    # comma separated (int)] :
    # If set to NULL, check if the TAG[MODEL_MAPPING_KERNEL] FITS BINTABLE values
    # is provided.
    # The FITS BINTABLE have to one column [KERNEL_LIBRARY_EXT].
    fileout.write('MODEL_MAPPING_KERNEL=NULL\n')


    ## MOLECULAR COLUMNS

    # --LIST_MOLEC
    # List of molecules to be included in the model. Represented as a comma
    # separated
    # string of molecule names, e.g. "H2O,CO2,O3".
    # If set to NULL, the input TAG[MOLECULES] FITS BINTABLE values have to be
    # provided
    # where the FITS BINTABLE specified contains the three columns:
    # LIST_MOLEC; FIT_MOLEC; and REL_COL.
    molecules_list = 'LIST_MOLEC="'

    # --FIT_MOLEC
    # List of flags that specify which of the listed molecules are to be fitted for.
    # Flag=1 implies yes. Flag=0 implies no. Represented as a string of comma
    # separated
    # integers in the same order as the listed molecules. For example: if
    # LIST_MOLEC="H2O,CO2,O3", then
    # FIT_MOLEC="1,0,1" implies that only H2O and O3 should be fitted for.
    # If set to NULL, the input TAG[MOLECULES] FITS BINTABLE values have to be
    # provided where the FITS
    # BINTABLE specified contains the three columns: LIST_MOLEC; FIT_MOLEC; and
    # REL_COL.
    molecules_flag = 'FIT_MOLEC="'

    # --REL_COL
    # List of the intial values of fitting of the molecular columns expressed
    # relatively to the input
    #  ATM profile columns. Represented as a comma separated list of doubles in
    # the same order as the
    # listed molecules. For example, if LIST_MOLEC="H2O,CO2,O3", then
    # REL_COL="1.0,1.2,0.8"
    # implies that H2O, CO2 and O3 have initial relative values of 1.0, 1.2 and
    # 0.8 respectively.
    # If set to NULL, the input TAG[MOLECULES] FITS BINTABLE values have to be
    # provided where the FITS
    # BINTABLE specified contains the three columns: LIST_MOLEC; FIT_MOLEC; and
    # REL_COL.
    molecules_rel = 'REL_COL="'
    for mol in molecfit_dict['molecules']:
        molecules_list += mol
        molecules_list += ','
        molecules_flag += '1,'
        molecules_rel += '1.0,'

    molecules_list = molecules_list[:-1] + '"'
    molecules_flag = molecules_flag[:-1] + '"'
    molecules_rel = molecules_rel[:-1] + '"'

    fileout.write(molecules_list + '\n')
    fileout.write(molecules_flag + '\n')
    fileout.write(molecules_rel + '\n')

    # --WAVE_INCLUDE
    # Wavelength ranges to be included. Represented as a string of comma separated
    # doubles in pairs
    # specifying the start and end wavelengths of a range. The wavelength units
    # are always in microns.
    # For example a KMOS sample data in the range of 1.11um to 1.67um may have
    # WAVE_INCLUDE="1.773,1.78633,1.79098,1.80434,1.187691,1.189937" to represent
    # three inclusion regions:
    # [1.773,1.78633], [1.79098,1.80434] and [1.187691,1.189937].
    # If set to NULL, molecfit will check if the TAG[WAVE_INCLUDE] FITS BINTABLE
    # values is provided where
    # the FITS BINTABLE specified has the two columns: LOWER_LIMIT; and
    # UPPER_LIMIT.
    fileout.write('WAVE_INCLUDE='+wave_include+'\n')

    # --WAVE_EXCLUDE
    # Wavelength ranges to be excluded. Represented as a string of comma separated
    # doubles in pairs
    # specifying the start and end wavelengths of a range. The wavelength units
    # are always in microns.
    # as the input science data. For example a KMOS sample data in the range of
    # 1.11um to 1.67um may have
    # WAVE_EXCLUDE="1.773,1.78633,1.79098,1.80434,1.187691,1.189937" to represent
    # three exclusion regions:
    # [1.773,1.78633], [1.79098,1.80434] and [1.187691,1.189937].
    # If set to NULL, molecfit will check if the TAG[WAVE_EXCLUDE] FITS BINTABLE
    # values is provided where
    # the FITS BINTABLE specified has the two columns: LOWER_LIMIT; and
    # UPPER_LIMIT.
    fileout.write('WAVE_EXCLUDE=NULL\n')


    # --PIXEL_EXCLUDE
    # Pixel ranges to be excluded. Represented as a string of comma separated
    # integers in pairs specifying the
    # start and end pixel of a range. For example:
    # PIXEL_EXCLUDE="54,128,512,514,1020,1024" represents three
    # exclusion regions: [54,128], [512,514] and [1020,1024].
    # If set to NULL, molecfit will check if the TAG[PIXEL_EXCLUDE] FITS BINTABLE
    # values is provided where the
    # FITS BINTABLE specified has the two columns: LOWER_LIMIT; and UPPER_LIMIT.
    fileout.write('PIXEL_EXCLUDE=NULL\n')

    # --TELLURICCORR_PATH
    # Installation directory.
    fileout.write('TELLURICCORR_PATH=TELLURICCORR_PARAMETER_DEFAULT\n')

    # --TELLURICCORR_DATA_PATH
    # Data directory.
    fileout.write('TELLURICCORR_DATA_PATH=TELLURICCORR_PARAMETER_DEFAULT\n')

    # --TMP_PATH
    # Temporary directory.
    fileout.write('TMP_PATH=TELLURICCORR_PARAMETER_DEFAULT\n')

    # --SILENT_EXTERNAL_BINS
    # Silent the output of the external binaries.
    fileout.write('SILENT_EXTERNAL_BINS=TRUE\n')

    # --TRANSMISSION
    # Type of input spectrum : 0 = Emission(radiance); 1 = Transmission.
    fileout.write('TRANSMISSION=TRUE\n')

    # --COLUMN_LAMBDA
    # Wavelength column ('NULL' can be used if the file is an image and that
    # the data are in the primary
    # (data are given by the FITS header keywords [CRVAL1=wave_ini, CD1_1=step])
    # f CD1_1 is absent, then the DEPRECATED CDELT1 keyword will be used.
    fileout.write('COLUMN_LAMBDA=lambda\n')

    # --COLUMN_FLUX
    # Flux column.
    fileout.write('COLUMN_FLUX=flux\n')

    # --COLUMN_DFLUX
    # Flux error column (Avoided by writing 'NULL') : 1-sigma error on the flux.
    fileout.write('COLUMN_DFLUX=NULL\n')

    # --COLUMN_MASK
    # Mask column (Avoided by writing 'NULL') : Indicates if a pixel is invalid.
    fileout.write('COLUMN_MASK=NULL\n')

    # --DEFAULT_ERROR
    # Default error relative to mean for the case that the error column
    # is not provided.
    fileout.write('DEFAULT_ERROR=0.01\n')

    # --WLG_TO_MICRON
    # Multiplicative factor applied to the wavelength to express is in micron.
    # E.g.: if wavelength is given in nm, the value should be 0.001.
    fileout.write('WLG_TO_MICRON=1.\n')

    # --WAVELENGTH_FRAME
    # Wavelength in vacuum                                      = 'VAC'.
    # Wavelength in air    with the observatory reference frame = 'AIR'.
    # Wavelength in vacuum with another         reference frame = 'VAC_RV'.
    #   (typically the  sun or the barycenter of the solar system).
    # In the latter case, the radial velocity of the observatory relative
    #   to the external reference frame must be provided in the parameter obs_RV.
    fileout.write('WAVELENGTH_FRAME=AIR\n')
    # TODO should I convert everything in vacuum or air??


    # --OBS_ERF_RV_KEY
    # The radial velocity of the observatory in km/s
    # relative to the external reference frame;
    # It is positive if the distance between the science target and the Earth
    # increases along the line-of-sight to the science target.
    # It must be provided if MF_PARAMETERS_WAVELENGTH_FRAME = 'VAC_RV'.
    fileout.write('OBS_ERF_RV_KEY=NONE\n')

    # --OBS_ERF_RV_VALUE
    # If OBS_ERF_RV_KEYWORD=='NONE' take this value.
    fileout.write('OBS_ERF_RV_VALUE=0.0\n')

    # --CLEAN_MODEL_FLUX
    # Set model flux to 0 for non-fitted pixels.
    fileout.write('CLEAN_MODEL_FLUX=FALSE\n')

    # --FTOL
    # Relative chi-square convergence criterion.
    # FTOL=1e-10
    try:
        fileout.write('FTOL={0:.12f}\n'.format(molecfit_dict['ftol']))
    except ValueError:
        fileout.write("FTOL=" + molecfit_dict['ftol'] + "\n")

    # --XTOL
    # Relative parameter convergence criterion.
    # XTOL=1e-10
    try:
        fileout.write('XTOL={0:.12f}\n'.format(molecfit_dict['xtol']))
    # Relative chi2 convergence criterion
    except ValueError:
        fileout.write("XTOL=" + molecfit_dict['xtol'] + "\n")

    # --FLUX_UNIT
    # Conversion of fluxes from phot/(s*m2*mum*as2) (emission spectrum only)
    #    to flux unit of observed spectrum:
    # 0: phot / (s *  m^2 * mum * as^2) [no conversion]
    # 1:    W / (     m^2 * mum * as^2)
    # 2:  erg / (s * cm^2 *   A * as^2)
    # 3:  mJy / (                 as^2)
    # For other units, the conversion factor has to be considered
    #  as constant term of the continuum fit.
    fileout.write('FLUX_UNIT=0\n')

    # --FIT_TELESCOPE_BACKGROUND
    # Fit of telescope background -- 1 = yes; 0 = no (emission spectrum only).
    fileout.write('FIT_TELESCOPE_BACKGROUND=TRUE\n')
    #todo check this, it was zero in previous file format

    # --TELESCOPE_BACKGROUND_CONST
    # Initial value for telescope background fit.
    fileout.write('TELESCOPE_BACKGROUND_CONST=0.1\n')

    # --FIT_CONTINUUM
    # Comma deliminated string of flags (1=true, 0=false) for fitting continuum in
    # specific regions.
    # If set to NULL, check if the TAG[WAVE_INCLUDE] points to a FITS BINTABLE
    # with column CONT_FIT_FLAG provided.
    fileout.write('FIT_CONTINUUM=1\n')

    # --CONTINUUM_N
    # Polynomial order for continuum model for each region. Presented as a comma
    # deliminated string.
    # If set to NULL, check if the TAG[WAVE_INCLUDE] points to a FITS BINTABLE
    # with column CONT_POLY_ORDER provided.
    fileout.write('CONTINUUM_N={0:1.0f}\n'.format(molecfit_dict['cont_n']))

    # --CONTINUUM_CONST
    # Initial constant term for continuum fit (valid for all fit ranges)
    # [emission spectrum: about 1 for correct flux_unit].
    fileout.write('CONTINUUM_CONST={0:10f}\n'.format(molecfit_dict['cont_const']))

    # --FIT_WLC
    # Flags for including regions in wavelength corrections.
    fileout.write('FIT_WLC=1\n')

    # --WLC_N
    # Polynomial degree of the refined wavelength solution.
    fileout.write('WLC_N={0:1.0f}\n'.format(molecfit_dict['wlc_n']))

    # --WLC_CONST
    # Initial constant term for wavelength adjustment
    # (shift relative to half wavelength range).
    fileout.write('WLC_CONST={0:10f}\n'.format(molecfit_dict['wlc_const']))

    # --FIT_RES_BOX
    # Fit resolution by Boxcar LSF.
    fileout.write('FIT_RES_BOX=FALSE\n')

    # --RES_BOX
    # Initial value for FWHM of Boxcar rel. to slit width
    # at the centre of the spectrum.
    fileout.write('RES_BOX=0.0\n')

    # --FIT_RES_GAUSS
    # Fit resolution by Gaussian.
    fileout.write('FIT_RES_GAUSS=TRUE\n')


    # --RES_GAUSS
    # Initial value for FWHM of the Gaussian in pixels
    # at the centre of the spectrum.
    fileout.write('RES_GAUSS={0:10f}\n'.format(molecfit_dict['res_gauss']))

    # --FIT_RES_LORENTZ
    # Fit resolution by Lorentzian.
    fileout.write('FIT_RES_LORENTZ=FALSE\n')

    # --RES_LORENTZ
    # Initial value for FWHM of the Lorentz in pixels
    # at the centre of the spectrum.
    fileout.write('RES_LORENTZ=0.0\n')

    # --KERNMODE
    # Voigtian profile approximation instead of independent Gaussian and
    # Lorentzian?.
    fileout.write('KERNMODE=FALSE\n')

    # --KERNFAC
    # Size of Voigtian/Gaussian/Lorentzian kernel in FWHM.
    fileout.write('KERNFAC={0:10f}\n'.format(molecfit_dict['kernfac']))

    # --VARKERN
    # Does the kernel size increase linearly with wavelength?.
    fileout.write('VARKERN=TRUE\n')
    #TODO: check why it was FALSE in the previous iteration

    # --OBSERVING_DATE_KEYWORD
    # Observing date in years or MJD in days (not string).
    fileout.write('OBSERVING_DATE_KEYWORD=NONE\n')

    # --OBSERVING_DATE_VALUE
    # If OBSERVING_DATE_KEYWORD=='NONE' take this value.
    fileout.write('OBSERVING_DATE_VALUE={0:30f}\n'.format(observing_dict['MJD']))

    # --UTC_KEYWORD
    # UTC in s.
    fileout.write('UTC_KEYWORD=NONE\n')

    # --UTC_VALUE
    # If UTC_KEYWORD=='NONE' take this value.
    fileout.write('UTC_VALUE={0:30f}\n'.format(observing_dict['UTC']))

    # --TELESCOPE_ANGLE_KEYWORD
    # Telescope altitude angle in deg.
    fileout.write('TELESCOPE_ANGLE_KEYWORD=NONE\n')

    # --TELESCOPE_ANGLE_VALUE
    # If TELESCOPE_ANGLE_KEYWORD=='NONE' take this value.
    fileout.write('TELESCOPE_ANGLE_VALUE={0:30f}\n'.format(observing_dict['ELEVATION']))

    # --RELATIVE_HUMIDITY_KEYWORD
    # Relative humidity in %.
    fileout.write('RELATIVE_HUMIDITY_KEYWORD=NONE\n')

    # --RELATIVE_HUMIDITY_VALUE
    # If RELATIVE_HUMIDITY_KEYWORD=='NONE' take this value.
    fileout.write('RELATIVE_HUMIDITY_VALUE={0:30f}\n'.format(observing_dict['HUMIDITY']))

    # --PRESSURE_KEYWORD
    # Pressure in hPa.
    fileout.write('PRESSURE_KEYWORD=NONE\n')

    # --PRESSURE_VALUE
    # If PRESSURE_KEYWORD=='NONE' take this value.
    fileout.write('PRESSURE_VALUE={0:5.1f}\n'.format(observing_dict['PRESSURE']))

    # --TEMPERATURE_KEYWORD
    # Ambient temperature in deg C.
    fileout.write('TEMPERATURE_KEYWORD=NONE\n')

    # --TEMPERATURE_VALUE
    # If TEMPERATURE_KEYWORD=='NONE' take this value.
    fileout.write('TEMPERATURE_VALUE={0:4.1f}\n'.format(observing_dict['TEMPERATURE_EN']))

    # --MIRROR_TEMPERATURE_KEYWORD
    # Mirror temperature in deg C.
    fileout.write('MIRROR_TEMPERATURE_KEYWORD=NONE\n')

    # --MIRROR_TEMPERATURE_VALUE
    # If MIRROR_TEMPERATURE_KEYWORD=='NONE' take this value.
    fileout.write('MIRROR_TEMPERATURE_VALUE={0:4.1f}\n'.format(observing_dict['TEMPERATURE_M1']))

    # --ELEVATION_KEYWORD
    # Elevation above sea level in m (default is Paranal: 2635. m).
    fileout.write('ELEVATION_KEYWORD=NONE\n')

    # --ELEVATION_VALUE
    # If ELEVATION_KEYWORD=='NONE' take this value.
    fileout.write('ELEVATION_VALUE={0:30f}\n'.format(observing_dict['GEOELEV']))

    # --LONGITUDE_KEYWORD
    # Longitude (default is Paranal: -70.4051 deg).
    fileout.write('LONGITUDE_KEYWORD=NONE\n')

    # --LONGITUDE_VALUE
    # If LONGITUDE_KEYWORD=='NONE' take this value.
    fileout.write('LONGITUDE_VALUE={0:9.4f}\n'.format(observing_dict['GEOLONG']))

    # --LATITUDE_KEYWORD
    # Latitude (default is Paranal: -24.6276 deg).
    fileout.write('LATITUDE_KEYWORD=NONE\n')

    # --LATITUDE_VALUE
    # If LATITUDE_KEYWORD=='NONE' take this value.
    fileout.write('LATITUDE_VALUE={0:9.4f}\n'.format(observing_dict['GEOLAT']))

    # --SLIT_WIDTH_KEYWORD
    # Slit width in arcsec (taken from FITS header if present).
    fileout.write('SLIT_WIDTH_KEYWORD=NONE\n')

    # --SLIT_WIDTH_VALUE
    # If SLIT_WIDTH_KEYWORD=='NONE' take this value.
    fileout.write('SLIT_WIDTH_VALUE={0:3.1f}\n'.format(molecfit_dict['slitwidth']))

    # --PIX_SCALE_KEYWORD
    # Pixel scale in arcsec (taken from this file only).
    fileout.write('PIX_SCALE_KEYWORD=NONE\n')

    # --PIX_SCALE_VALUE
    # If PIX_SCALE_KEYWORD=='NONE' take this value.
    fileout.write('PIX_SCALE_VALUE={0:4.12}\n'.format(molecfit_dict['pixelscale']))

    # --REFERENCE_ATMOSPHERIC
    # Reference atmospheric profile. Possible values:
    # - equ_atm (default; equatorial atmosphere, valid for Paranal);
    # - tro_atm (tropical atmosphere);
    # - std_atm (standard atmosphere);
    # - Other file located in :
    #   ({TELLURICCORR_DATA_PATH}/profiles/mipas/).
    fileout.write('REFERENCE_ATMOSPHERIC=equ.fits\n')

    # --GDAS_PROFILE
    # Specific GDAS-like input profile (P[hPa] HGT[m] T[K] RELHUM[%])
    # (if starts with /, absolute path, otherwise path relative to basedir).
    # In the case of 'none', no GDAS profiles will be considered.
    # The value 'auto' performs an automatic retrieval.
    fileout.write('GDAS_PROFILE=auto\n')

    # --LAYERS
    # Grid of layer heights for merging ref_atm and GDAS profile.
    # Fixed grid = CPL_TRUE and natural grid = CPL_FALSE.
    fileout.write('LAYERS=TRUE\n')

    # --EMIX
    # Upper mixing height in km for considering data of a local meteo station.
    # If emix is below geoelev, rhum, pres, and temp are not used
    # for modifying the corresponding profiles.
    fileout.write('EMIX=5.0\n')

    # --PWV
    # PWV value in mm for the input water vapor profile.
    # The merged profile composed of ref_atm, GDAS, and local meteo data
    # will be scaled to this value if pwv > 0 (default: -1 -> no scaling).
    fileout.write('PWV=-1.0\n')

    # --LNFL_LINE_DB
    # File name of the line list (must be stored in the directory :
    # ({TELLURICCORR_DATA_PATH}/hitran/).
    fileout.write('LNFL_LINE_DB=aer_v_'+molecfit_dict['aer_version']+'\n')

    # --LNFL_LINE_DB_FORMAT
    # Format of the line file: gives the length in terms of characters per line.
    fileout.write('LNFL_LINE_DB_FORMAT=100.0\n')

    # --LBLRTM_ICNTNM
    # Continua and Rayleigh extinction [0,1,2,3,4,5].
    fileout.write('LBLRTM_ICNTNM=5\n')

    # --LBLRTM_IAERSL
    # Aerosols [0,1].
    fileout.write('LBLRTM_IAERSL=0\n')

    # --LBLRTM_MPTS
    # Number of optical depth values.
    fileout.write('LBLRTM_MPTS=5\n')

    # --LBLRTM_NPTS
    # Number of values for each panel.
    fileout.write('LBLRTM_NPTS=5\n')

    # --LBLRTM_V1
    # Beginning wavenumber value for the calculation.
    fileout.write('LBLRTM_V1=1.9\n')

    # --LBLRTM_V2
    # Ending wavenumber value for the calculation.
    fileout.write('LBLRTM_V2=2.4\n')

    # --LBLRTM_SAMPLE
    # Number of sample points per mean halfwidth [between 1 to 4, default=4].
    fileout.write('LBLRTM_SAMPLE=4\n')

    # --LBLRTM_ALFAL0
    # Average collision broadened halfwidth [cm-1/atm].
    fileout.write('LBLRTM_ALFAL0=0.0\n')

    # --LBLRTM_AVMASS
    # Average molecular mass [amu] for Doppler halfwidth.
    fileout.write('LBLRTM_AVMASS=0.0\n')

    # --LBLRTM_DPTMIN
    # Minimum molecular optical depth below which lines will be rejected.
    fileout.write('LBLRTM_DPTMIN=0.0002\n')

    # --LBLRTM_DPTFAC
    # Factor multiplying molecular continuum optical depth.
    fileout.write('LBLRTM_DPTFAC=0.001\n')

    # --LBLRTM_TBOUND
    # Temperature of boundary [K].
    fileout.write('LBLRTM_TBOUND=0.0\n')

    # --LBLRTM_SREMIS1
    # Emissivity coefficient 1.
    fileout.write('LBLRTM_SREMIS1=0.0\n')

    # --LBLRTM_SREMIS2
    # Emissivity coefficient 2.
    fileout.write('LBLRTM_SREMIS2=0.0\n')

    # --LBLRTM_SREMIS3
    # Emissivity coefficient 3.
    fileout.write('LBLRTM_SREMIS3=0.0\n')

    # --LBLRTM_SRREFL1
    # Reflectivity coefficient 1.
    fileout.write('LBLRTM_SRREFL1=0.0\n')

    # --LBLRTM_SRREFL2
    # Reflectivity coefficient 2.
    fileout.write('LBLRTM_SRREFL2=0.0\n')

    # --LBLRTM_SRREFL3
    # Reflectivity coefficient 3.
    fileout.write('LBLRTM_SRREFL3=0.0\n')

    # --LBLRTM_MODEL
    # Atmospheric profile [0,1,2,3,4,5,6].
    fileout.write('LBLRTM_MODEL=0\n')

    # --LBLRTM_ITYPE
    # Type of path [1,2,3].
    fileout.write('LBLRTM_ITYPE=3\n')

    # --LBLRTM_NOZERO
    # Zeroing of small amounts of absorbers [0,1].
    fileout.write('LBLRTM_NOZERO=0\n')

    # --LBLRTM_NOPRNT
    # Do not print output? [0,1].
    fileout.write('LBLRTM_NOPRNT=0\n')

    # --LBLRTM_IPUNCH
    # Write out layer data to TAPE7 [0,1].
    fileout.write('LBLRTM_IPUNCH=0\n')

    # --LBLRTM_RE
    # Radius of earth [km].
    fileout.write('LBLRTM_RE=0.0\n')

    # --LBLRTM_HSPACE
    # Altitude definition for space [km].
    fileout.write('LBLRTM_HSPACE=120.0\n')

    # --LBLRTM_H2
    # Upper height limit [km].
    fileout.write('LBLRTM_H2=0.0\n')

    # --LBLRTM_RANGE
    # Length of a straight path from H1 to H2 [km].
    fileout.write('LBLRTM_RANGE=0.0\n')

    # --LBLRTM_BETA
    # Earth centered angle from H1 to H2 [degrees].
    fileout.write('LBLRTM_BETA=0.0\n')

    # --LBLRTM_LEN
    # Path length [0,1].
    fileout.write('LBLRTM_LEN=0\n')

    # --LBLRTM_HOBS
    # Height of observer.
    fileout.write('LBLRTM_HOBS=0.0\n')

    # --LBLRTM_AVTRAT
    # Maximum Voigt width ratio across a layer.
    fileout.write('LBLRTM_AVTRAT=2.0\n')

    # --LBLRTM_TDIFF1
    # Maximum layer temperature difference at ALTD1 [K].
    fileout.write('LBLRTM_TDIFF1=5.0\n')

    # --LBLRTM_TDIFF2
    # Maximum layer temperature difference at ALTD2 [K].
    fileout.write('LBLRTM_TDIFF2=8.0\n')

    # --LBLRTM_ALTD1
    # Altitude of TDIFF1 [km].
    fileout.write('LBLRTM_ALTD1=0.0\n')

    # --LBLRTM_ALTD2
    # Altitude of TDIFF2 [km].
    fileout.write('LBLRTM_ALTD2=0.0\n')

    # --LBLRTM_DELV
    # Number of wavenumbers [cm-1] per major division.
    fileout.write('LBLRTM_DELV=1.0\n')

    # --EXPERT_MODE
    # If set to true, will check if TAG[INIT_FIT_PARAMETERS] points to a fits file
    # with a bintable of parameter values to use as initial values for the
    # fitting process.
    fileout.write('EXPERT_MODE=FALSE\n')

    # --CHIP_EXTENSIONS
    # Flag that determines if image extensions are to be treated as independent
    # science data to be fitted for independently or as CHIP specific subranges
    # of a single observation to be fitted for as a single combined spectrum.
    # Value = TRUE implies to treat as CHIPS to be combined. Value = FALSE
    # implies
    # to treat as independent. [FALSE].
    fileout.write('CHIP_EXTENSIONS=FALSE\n')

    #
    # End of file


    fileout.close()


def write_molecfit_input_spectrum(wave, flux, filename, keywords=None):
    #writing the molecfit input fits files

    good=range(len(wave))

    for i in range(5):

        medFlux=np.median(flux[good])
        madFlux=np.median(np.abs(flux[good]-medFlux))
        good=np.where(flux>(medFlux-5*madFlux))[0]

    dflux=np.std(flux[good])
    dflux=dflux*np.sqrt(flux)#to scale the error to the signal assuming poissonian noise

    #To create a table from scratch, we need to define columns first, by constructing the Column objects and their data.
    #Suppose we have two columns, the first containing strings, and the second containing floating point numbers:

    col1 = fits.Column(name='lambda', format='E', array=wave*1e-4)
    col2 = fits.Column(name='flux', format='E', array=flux)
    col3 = fits.Column(name='dflux', format='E', array=dflux)
    col4 = fits.Column(name='mask', format='I', array=np.zeros(len(good)))

    #Now, create a new binary table HDU object by using the BinTableHDU.from_columns() function:
    #hdu = fits.BinTableHDU.from_columns(cols)
    table_hdu = fits.BinTableHDU.from_columns([col1, col2, col3, col4])

    #The data structure used to represent FITS tables is called a FITS_rec and is derived from the numpy.recarray interface.
    #When creating a new table HDU the individual column arrays will be assembled into a single FITS_rec array.

    if keywords is None:
        #Now you may write this new table HDU directly to a FITS file like so:
        table_hdu.writeto(filename, overwrite=True)
    else:
        primary_hdu = fits.PrimaryHDU()
        for key, value in keywords.items():
            primary_hdu.header[key] = value
        hdu_out = fits.HDUList([primary_hdu, table_hdu])
        hdu_out.writeto(filename, overwrite=True)

def write_calctrans_par(filename_par):
    fileout = open(filename_par, 'w')


    # File: /archive/molecfitRootDataDir/calctrans.rc
    #
    # Note: This configuration file has been automatically
    #       generated by the esorex (v3.13.3) program.
    #
    # Date: 17-Jan-2021 18:30:01
    #
    #

    # --USE_ONLY_INPUT_PRIMARY_DATA
    # Value=TRUE implies that only the fits primary contains the input science flux
    # data.
    # Value=FALSE implies that the fits extensions also contains input science
    # flux data.
    fileout.write('USE_ONLY_INPUT_PRIMARY_DATA=FALSE\n')

    # --USE_DATA_EXTENSION_AS_DFLUX
    # Only valid if USE_ONLY_INPUT_PRIMARY_DATA=TRUE. The fits extension index that
    # contains the
    # errors of the science flux data (DFLUX). A value of 0 implies that there is
    # no DFLUX.
    fileout.write('USE_DATA_EXTENSION_AS_DFLUX=0\n')

    # --USE_DATA_EXTENSION_AS_MASK
    # Only valid if USE_ONLY_INPUT_PRIMARY_DATA=TRUE. The fits extension index that
    # contains the
    # mask associated with the science flux data. A value of 0 implies that there
    # is no mask data.
    fileout.write('USE_DATA_EXTENSION_AS_MASK=0\n')

    # --USE_INPUT_KERNEL
    # If TRUE, use the kernel library if it is provided.
    fileout.write('USE_INPUT_KERNEL=TRUE\n')

    # --CALCTRANS_MAPPING_KERNEL
    # Mapping 'SCIENCE' - 'CALCTRANS_KERNEL_LIBRARY' [string with ext_number comma
    # separated (int)] :
    # If set to NULL, check if the TAG[CALCTRANS_MAPPING_KERNEL] FITS BINTABLE
    # values is provided.
    # The FITS BINTABLE have to one column [KERNEL_LIBRARY_EXT].
    fileout.write('CALCTRANS_MAPPING_KERNEL=NULL\n')

    # --MAPPING_ATMOSPHERIC
    # Mapping 'SCIENCE' - 'ATM_PARAMETERS' input [string with ext_number comma
    # separated (int)] :
    # If set to NULL, check if the TAG[MAPPING_ATMOSPHERIC] FITS BINTABLE value
    # is provided.
    # The FITS BINTABLE have to one column [ATM_PARAMETERS_EXT].
    fileout.write('MAPPING_ATMOSPHERIC=0,1\n')

    # --MAPPING_CONVOLVE
    # Mapping 'LBLRTM_RESULTS' - 'TELLURIC_CORR' output [string with ext_number
    # comma separated (int)] :
    # If set to NULL, check if the TAG[MAPPING_CONVOLVE] FITS BINTABLE value is
    # provided.
    # The FITS BINTABLE have to one column [LBLRTM_RESULTS_EXT].
    fileout.write('MAPPING_CONVOLVE=0,1\n')

    # --CHIP_EXTENSIONS
    # Flag that determines if image extensions are to be treated as independant
    # science data to be fitted for independently or as CHIP specific subranges
    # of a single observation to be fitted for as a single combined spectrum.
    # Value = TRUE implies to treat as CHIPS to be combined. Value = FALSE
    # implies
    # to treat as independent. [FALSE].
    fileout.write('CHIP_EXTENSIONS=FALSE\n')

    #
    # End of file

    fileout.close()

def write_correct_par(filename_par):

    fileout = open(filename_par, 'w')

    # File: /archive/molecfitRootDataDir/correct.rc
    #
    # Note: This configuration file has been automatically
    #       generated by the esorex (v3.13.3) program.
    #
    # Date: 17-Jan-2021 18:30:12
    #
    #

    # --USE_ONLY_INPUT_PRIMARY_DATA
    # Value=TRUE implies that only the fits primary contains the input science flux
    # data.
    # Value=FALSE implies that the fits extensions also contains input science
    # flux data.
    fileout.write('USE_ONLY_INPUT_PRIMARY_DATA=FALSE\n')

    # --USE_DATA_EXTENSION_AS_DFLUX
    # Only valid if USE_ONLY_INPUT_PRIMARY_DATA=TRUE. The fits extension index that
    # contains the
    # errors of the science flux data (DFLUX). A value of 0 implies that there is
    # no DFLUX.
    fileout.write('USE_DATA_EXTENSION_AS_DFLUX=0\n')

    # --USE_DATA_EXTENSION_AS_MASK
    # Only valid if USE_ONLY_INPUT_PRIMARY_DATA=TRUE. The fits extension index that
    # contains the
    # mask associated with the science flux data. A value of 0 implies that there
    # is no mask data.
    fileout.write('USE_DATA_EXTENSION_AS_MASK=0\n')

    # --SUPPRESS_EXTENSION
    # Suppress arbitrary filename extension : TRUE (apply) or FALSE (don't apply).
    fileout.write('SUPPRESS_EXTENSION=FALSE\n')

    # --MAPPING_CORRECT
    # Mapping 'SCIENCE' - 'TELLURIC_CORR' [string with ext_number comma separated
    # (int)] :
    # If set to NULL, check if the TAG[MAPPING_CORRECT] FITS BINTABLE value is
    # provided.
    # The FITS BINTABLE have to one column [TELLURIC_CORR_EXT].
    fileout.write('MAPPING_CORRECT=0,1\n')

    # --CHIP_EXTENSIONS
    # Flag that determines if image extensions are to be treated as independant
    # science data to be fitted for independently or as CHIP specific subranges
    # of a single observation to be fitted for as a single combined spectrum.
    # Value = TRUE implies to treat as CHIPS to be combined. Value = FALSE
    # implies
    # to treat as independent. [FALSE].
    fileout.write('CHIP_EXTENSIONS=FALSE\n')

    #
    # End of file

    fileout.close()



from __future__ import print_function, division
from SLOPpy.subroutines.common import *
from SLOPpy.subroutines.spectral_subroutines import *
from SLOPpy.subroutines.io_subroutines import *
from SLOPpy.subroutines.fit_subroutines import *
from SLOPpy.subroutines.shortcuts import *
from SLOPpy.subroutines.plot_subroutines import *

from scipy.interpolate import UnivariateSpline
from scipy.signal import savgol_filter


__all__ = ['compute_transmission_spectrum_preparation',
           'plot_transmission_spectrum_preparation']

def compute_transmission_spectrum_preparation(config_in):

    subroutine_name = 'transmission_spectrum_preparation'

    night_dict = from_config_get_nights(config_in)

    for night in night_dict:

        try:
            preparation = load_from_cpickle('transmission_preparation',
                                          config_in['output'],
                                          night)
            print("{0:45s} Night:{1:15s}   {2:s}".format(subroutine_name, night, 'Retrieved'))
            continue
        except:
            print("{0:45s} Night:{1:15s}   {2:s}".format(subroutine_name, night, 'Computing'))
            print()

        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)

        calib_data = load_from_cpickle('calibration_fibA', config_in['output'], night)
        input_data = retrieve_observations(config_in['output'], night, lists['observations'])
        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

        if config_in['master-out'].get('use_composite', False):
            master_out = load_from_cpickle('master_out_composite', config_in['output'], night)
            print('  Using composite master-out from all nights')
        else:
            master_out = load_from_cpickle('master_out', config_in['output'], night)

        if config_in['master-out'].get('use_smoothed', False):
            master_out['rescaled'] = master_out['smoothed']
            master_out['rescaled_err'] = master_out['smoothed_err']
            print('  Using smoothed master-out')

        preparation = {
            'subroutine': subroutine_name,
        }

        for obs in lists['observations']:

            preparation[obs] = {}

            preparation[obs]['master_out'] = {}
            preparation[obs]['wave'] = input_data[obs]['wave'] #Added for plotting purpose only

            """ Step 1+2): bring back the master-out to the ORF and rebin the 1D master-out to the 2D observation scale"""
            preparation[obs]['master_out']['rebinned'] = \
                rebin_1d_to_2d(master_out['wave'],
                               master_out['step'],
                               master_out['rescaled'],
                               input_data[obs]['wave'],
                               input_data[obs]['step'],
                               rv_shift=-observational_pams[obs]['rv_shift_ORF2SRF_mod'],
                               preserve_flux=False)

            preparation[obs]['master_out']['rebinned_err'] = \
                rebin_1d_to_2d(master_out['wave'],
                               master_out['step'],
                               master_out['rescaled_err'],
                               input_data[obs]['wave'],
                               input_data[obs]['step'],
                               rv_shift=-observational_pams[obs]['rv_shift_ORF2SRF_mod'],
                               preserve_flux=False,
                               is_error=True)

            for order in range(0, observational_pams['n_orders']):
                preparation[obs]['master_out']['rebinned'][order, :], \
                preparation[obs]['master_out']['rebinned_err'][order, :], \
                _ = \
                    replace_values_errors_with_interpolation_1d(preparation[obs]['master_out']['rebinned'][order, :],
                                                                preparation[obs]['master_out']['rebinned_err'][order, :],
                                                                less_than=0.001, greater_than=5.0000)

            #replace_values_errors(preparation[obs]['master_out']['rebinned'],
            #                      preparation[obs]['master_out']['rebinned_err'],
            #                      threshold=0.0001, replacement=1.0000)

            """ Step 3): obtain the unscaled transmission spectrum for this observation """
            preparation[obs]['ratio'] = input_data[obs]['e2ds']/\
                                      preparation[obs]['master_out']['rebinned']
            preparation[obs]['ratio_err'] = preparation[obs]['ratio'] * \
                                          np.sqrt((input_data[obs]['e2ds_err']/
                                                   input_data[obs]['e2ds'])**2 +
                                                  (preparation[obs]['master_out']['rebinned_err']/
                                                   preparation[obs]['master_out']['rebinned'])**2)

            preparation[obs]['ratio_precleaning'] = preparation[obs]['ratio'].copy()
            preparation[obs]['ratio_precleaning_err'] = preparation[obs]['ratio_err'].copy()


        if night_dict[night].get('spline_residuals', True):

            print()
            print('   Cleaning for telluric residuals with Univariate Spline - threshold about 5%')
            # cleaning using spline_univariate
            for order in range(0, observational_pams['n_orders']):
                obs_reference =  lists['observations'][0]

                len_y = len(lists['observations'])
                len_x = len(preparation[obs_reference]['wave'][order, :])

                time_from_transit = np.empty(len_y, dtype=np.double)
                data_array = np.empty([len_y, len_x], dtype=np.double)
                median_array =  np.empty(len_y, dtype=np.double)
                for i_obs, obs in enumerate(lists['observations']):
                    time_from_transit[i_obs] =  input_data[obs]['BJD'] - observational_pams['time_of_transit']
                    median_array[i_obs] = np.median(preparation[obs]['ratio_precleaning'][order ,:])
                    data_array[i_obs, :] = preparation[obs]['ratio_precleaning'][order ,:]/median_array[i_obs]
                    #wave = preparation[obs]['wave'][order, :]

                res = data_array * 1.
                val = np.empty([len_y, len_x], dtype=np.double)

                for ii in range(0, len_x):
                    spl = UnivariateSpline(time_from_transit, data_array[:, ii])
                    val[:,ii] = spl(time_from_transit)
                    res[:,ii] -= val[:,ii]
                    res[:,ii] /= val[:,ii]

                sel = np.abs(res) > 0.05

                for i_obs, obs in enumerate(lists['observations']):
                    if np.sum(sel[i_obs]) > 0:
                        preparation[obs]['ratio'][order, sel[i_obs]] = val[i_obs, sel[i_obs]] * median_array[i_obs]
                        preparation[obs]['ratio_err'][order, sel[i_obs]] *= 10.
        else:
            print()
            print('   Cleaning for telluric residuals NOT performed')


        for obs in lists['observations']:

            preparation[obs]['deblazed'] = preparation[obs]['ratio'] / calib_data['blaze'] / (input_data[obs]['step'] / np.median(input_data[obs]['step']))
            preparation[obs]['deblazed_err'] = preparation[obs]['ratio_err'] / calib_data['blaze'] / (input_data[obs]['step'] / np.median(input_data[obs]['step']))


            if not config_in['settings'].get('full_output', False):
                del preparation[obs]['master_out']
            else:
                # added for plotting purposes only
                preparation[obs]['rescaling'], \
                preparation[obs]['rescaled'], \
                preparation[obs]['rescaled_err'] = perform_rescaling(
                    preparation[obs]['wave'],
                    preparation[obs]['deblazed'],
                    preparation[obs]['deblazed_err'],
                    observational_pams['wavelength_rescaling'])

        save_to_cpickle('transmission_preparation', preparation, config_in['output'], night)

    print()
    """ Keep going from here after preparation, unless the subroutines has been called just
        to preform the data preparation step
    """


def plot_transmission_spectrum_preparation(config_in, night_input=''):

    subroutine_name = 'transmission_spectrum_preparation'

    night_dict = from_config_get_nights(config_in)

    if night_input == '':
        night_list = night_dict
    else:
        night_list = np.atleast_1d(night_input)




    for night in night_list:

        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)
        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

        # ! To be removed when testing is done
        # ! This plots do not make any sense anymore
        input_data = retrieve_observations(config_in['output'], night, lists['observations'])


        """ Retrieving the analysis"""
        try:
            preparation = load_from_cpickle('transmission_preparation', config_in['output'], night)
        except:
            print("No transmission spectrum results, no plots")
            print()
            continue


        #from SLOPpy.subroutines.lines_fit_functions import logprob_case12
        from matplotlib.colors import BoundaryNorm
        from matplotlib.ticker import MaxNLocator

        len_y = len(lists['observations'])
        len_x = 4096
        order= 11

        time_from_transit = np.empty(len_y, dtype=np.double)
        plot_data = np.empty([len_y, len_x], dtype=np.double)

        for i_obs, obs in enumerate(lists['observations']):
            time_from_transit[i_obs] =  input_data[obs]['BJD'] - observational_pams['time_of_transit']
            plot_data[i_obs, :] = preparation[obs]['deblazed'][order ,:]/ np.median(preparation[obs]['deblazed'][order ,:])
            wave = preparation[obs]['wave'][order, :]


        wave_meshgrid, time_meshgrid = np.meshgrid(wave, time_from_transit)

        cmap = plt.get_cmap('coolwarm')

        #levels = MaxNLocator(nbins=15).tick_values(
        #    plot_data.min(), plot_data.max())
        levels = MaxNLocator(nbins=21).tick_values(0.90, 1.10)
        norm = BoundaryNorm(levels, ncolors=cmap.N, clip=True)

        plt.figure(figsize=(15, 10))
        plt.title('Transmission map in observer reference frame\n {0:s}'.format(night))

        PCF = plt.contourf(wave_meshgrid, time_meshgrid,
                            plot_data, levels=levels, cmap=cmap)
        cbar = plt.colorbar(PCF)
        cbar.ax.set_ylabel('Intensity')
        plt.show()


        if night_dict[night].get('spline_residuals', True):
            res = plot_data * 1.
            from scipy.interpolate import UnivariateSpline
            for ii in range(0,4096):
                spl = UnivariateSpline(time_from_transit, plot_data[:, ii])
                val = spl(time_from_transit)
                res[:,ii] -= val
                res[:,ii] /= val



            cmap = plt.get_cmap('coolwarm')

            levels = MaxNLocator(nbins=10).tick_values(-0.05, 0.05)
            norm = BoundaryNorm(levels, ncolors=cmap.N, clip=True)
            plt.figure(figsize=(15, 10))
            plt.title('Residuals after dividing by UnivariateSpline Spline\n {0:s}'.format(night))

            PCF = plt.contourf(wave_meshgrid, time_meshgrid,
                                res, levels=levels, cmap=cmap)
            cbar = plt.colorbar(PCF)
            cbar.ax.set_ylabel('Intensity')
            plt.show()





        """ Creation of the color array, based on the BJD of the observations
        """
        colors_properties, colors_plot, colors_scatter = make_color_array_matplotlib3(lists, observational_pams)

        fig = plt.figure(figsize=(12, 6))

        gs = GridSpec(1, 2, width_ratios=[50, 1])
        ax1 = plt.subplot(gs[0, 0])

        ax1.set_ylim(0.90, 1.10)
        #ax2 = plt.subplot(gs[1, 0], sharex=ax1, sharey=ax1)
        cbax1 = plt.subplot(gs[:, 1])

        for obs in lists['transit_in']:

            #preparation[obs]['rescaling'], \
            #preparation[obs]['rescaled'], \
            #preparation[obs]['rescaled_err'] = perform_rescaling(
            #    preparation[obs]['wave'],
            #    preparation[obs]['deblazed'] / (input_data[obs]['step'] / np.median(input_data[obs]['step'])),
            #    preparation[obs]['deblazed_err'] / (input_data[obs]['step'] / np.median(input_data[obs]['step'])),
            #    observational_pams['wavelength_rescaling'])

            preparation[obs]['rescaling'], \
            preparation[obs]['rescaled'], \
            preparation[obs]['rescaled_err'] = perform_rescaling(
                preparation[obs]['wave'],
                preparation[obs]['deblazed'],
                preparation[obs]['deblazed_err'],
                observational_pams['wavelength_rescaling'])


            ax1.scatter(preparation[obs]['wave'],
                    preparation[obs]['rescaled'],
                    s=1, alpha=0.25,
                    color=colors_plot['mBJD'][obs])

        sm = plt.cm.ScalarMappable(cmap=colors_properties['cmap'], norm=colors_properties['norm']['mBJD'])
        sm.set_array([])  # You have to set a dummy-array for this to work...
        cbar = plt.colorbar(sm, cax=cbax1)
        cbar.set_label('BJD - 2450000.0')
        fig.subplots_adjust(wspace=0.05, hspace=0.4)
        plt.show()
